"""This is the game module. Written in 2019 in Born by Carlos and Lisa."""

import logging
import os
from random import Random
import subprocess
import sys
import time
import math
from warnings import warn

from . import layout
from .exceptions import FatalException, NonFatalException, NoFoodWarning, PlayerTimeout
from .gamestate_filters import noiser, update_food_age, relocate_expired_food
from .layout import initial_positions, get_legal_positions
from .network import setup_controller, ZMQPublisher
from .team import make_team
from .viewer import ProgressViewer, AsciiViewer, ReplyToViewer, ReplayWriter, ResultPrinter

_logger = logging.getLogger(__name__)
_mswindows = (sys.platform == "win32")


### Global constants
# All constants that are currently not redefinable in setup_game

#: The points a team gets for killing another bot
KILL_POINTS = 5

#: The maximum distance between two bots before noise is applied
SIGHT_DISTANCE = 5

#: The radius for the uniform noise
NOISE_RADIUS = 5

#: The lifetime of food pellets in a shadow in turns
MAX_FOOD_AGE = 30

#: Food pellet shadow distance
SHADOW_DISTANCE = 3

class TkViewer:
    def __init__(self, *, address, controller, geometry=None, delay=None, stop_after=None, fullscreen=False):
        self.proc = self._run_external_viewer(address, controller, geometry=geometry, delay=delay, stop_after=stop_after, fullscreen=fullscreen)

    def _run_external_viewer(self, subscribe_sock, controller, geometry, delay, stop_after, fullscreen):
        # Something on OS X prevents Tk from running in a forked process.
        # Therefore we cannot use multiprocessing here. subprocess works, though.
        viewer_args = [ str(subscribe_sock) ]
        if controller:
            viewer_args += ["--controller-address", str(controller)]
        if geometry:
            viewer_args += ["--geometry", "{0}x{1}".format(*geometry)]
        if fullscreen:
            viewer_args += ["--fullscreen"]
        if delay:
            viewer_args += ["--delay", str(delay)]
        if stop_after is not None:
            viewer_args += ["--stop-after", str(stop_after)]

        tkviewer = 'pelita.scripts.pelita_tkviewer'
        external_call = [sys.executable,
                        '-m',
                        tkviewer] + viewer_args
        _logger.debug("Executing: %r", external_call)
        # os.setsid will keep the viewer from closing when the main process exits
        # a better solution might be to decouple the viewer from the main process
        if _mswindows:
            p = subprocess.Popen(external_call, creationflags=subprocess.CREATE_NEW_PROCESS_GROUP)
        else:
            p = subprocess.Popen(external_call, preexec_fn=os.setsid)
        return p

def controller_exit(state, await_action='play_step'):
    """Wait for the controller to receive a action from a viewer

    action can be 'exit' (return True), 'play_setup', 'set_initial' (return True)
    """

    if state['controller']:
        todo = state['controller'].await_action(await_action)
        if todo == 'exit':
            return True
        elif todo in ('play_step', 'set_initial'):
            return False

def run_game(team_specs, *, layout_dict, layout_name="", max_rounds=300,
             seed=None, allow_camping=False, error_limit=5, timeout_length=3,
             viewers=None, viewer_options=None, store_output=False,
             team_names=(None, None), team_infos=(None, None),
             allow_exceptions=False, print_result=True):
    """ Run a pelita match.

    Parameters
    ----------
    team_specs : list[team0, team1]
              a list specifying the two teams to play the match with. The list
              items team0 and team1 must be either
               - a function with signature move(bot, state) -> (x, y)
                 The function takes a bot object and a state and returns the next
                 position (x, y) of the bot.

               - the path to a Python module that defines at least a function
                 called 'move' (specified as above) and a string TEAM_NAME.

              When a team_spec is a function, a "local" game will be run, when
              they are paths to Python modules, a "remote" game will be run instead.
              See below in section Notes for more details about local and remote games.

    layout_dict : dict
               a dictionary representing a maze, as returned by pelita.layout.parse_layout

    layout_name : str
               a name for the layout (will be used in the UI).

    max_rounds : int
              The maximum number of rounds to play before the game is over. Default: 300.

    seed : int
        seed used to initialize the random number generator.

    error_limit : int
                   The limit of non fatal errors to reach for a team before the
                   game is over and the team is disqualified. Non fatal errors are
                   timeouts and returning an illegal move. Fatal errors are raising
                   Exceptions. An error_limit of 0 will disable the limit.
                   Default: 5.

    timeout_length : int or float
                  Time in seconds to wait for the move function (or for the remote
                  client) to return. After timeout_length seconds are elapsed a
                  non-fatal error is recorded for the team.

    viewers : list[viewer1, viewer2, ...]
           List of viewers to attach to the game. Implemented viewers: 'ascii',
           'progress', tk'. If None, no viewer is attached.

    viewer_options : do not use!

    store_output : False or str
                if store_output is a string it will be interpreted as a path to a
                directory where to store stdout and stderr for the client processes.
                It helps in debugging issues with the clients.
                In the special case of store_output==subprocess.DEVNULL, stdout of
                the remote clients will be suppressed.

    team_names : tuple(team_name_0, team_name_1)
              a tuple containing the team names. If not given, names will be taken
              from the team module TEAM_NAME variable or from the function name.

    team_info : tuple(team_info_0, team_info_1)
              a tuple containing additional team info.

    allow_exceptions : bool
                    when True, allow teams to raise Exceptions. This is especially
                    useful when running local games, where you typically want to
                    see Exceptions do debug them. When running remote games,
                    allow_exceptions should be False, so that the game can collect
                    the exceptions and cleanly create a game-over state if
                    needed.

    print_result : bool
                when True (default), print the result of the match on the command line

    Notes
    -----

    - remote games
        If teams_specs is a list of two Python modules, a "remote" game will be
        played. This means that the game will create one client subprocess for
        each module. The game communicates with the clients using the local network
        and sends/receives messages in JSON format. The game master will ask for the
        next position of a bot over the network, the clients will call their move
        function to get the next position and will return the position over the
        network. This setup allows for complete isolation of the clients, i.e. they
        can not see or influence each other directly.

    - local games
        If team_specs is a list of two functions, a "local" game will be played. This
        means that the game will get the next position of a bot by calling the
        corresponding move function directly. This mode is particularly useful to play
        many games in the background without a UI, because it skips all the network
        overhead of the remote games.

    """

    # allow exception will force exceptions in the clients to be raised.
    # This flag must be used when using run_game directly, like in tests or
    # in background games

    # we create the initial game state
    state = setup_game(team_specs, layout_dict=layout_dict,
                       layout_name=layout_name, max_rounds=max_rounds,
                       allow_camping=allow_camping,
                       error_limit=error_limit, timeout_length=timeout_length,
                       seed=seed, viewers=viewers,
                       viewer_options=viewer_options,
                       store_output=store_output, team_names=team_names,
                       team_infos=team_infos,
                       print_result=print_result)

    # Play the game until it is gameover.
    while not state.get('gameover'):

        # this is only needed if we have a controller, for example a viewer
        # this function call *blocks* until the viewer has replied
        if controller_exit(state):
            # if the controller asks us, we'll exit and stop playing
            break

        # play the next turn
        state = play_turn(state, allow_exceptions=allow_exceptions)

    return state


def setup_viewers(viewers=None, options=None, print_result=True):
    """ Returns a list of viewers from the given strings. """
    if viewers is None:
        viewers = []

    if options is None:
        options = {}

    zmq_publisher = None

    viewer_state = {
        'viewers': [],
        'procs': [],
        'controller': None
    }

    for viewer in viewers:
        if viewer == 'ascii':
            viewer_state['viewers'].append(AsciiViewer())
        elif viewer == 'progress':
            viewer_state['viewers'].append(ProgressViewer())
        elif len(viewer) == 2 and viewer[0] == 'reply-to':
            viewer_state['viewers'].append(ReplyToViewer(viewer[1]))
        elif len(viewer) == 2 and viewer[0] == 'write-replay-to':
            viewer_state['viewers'].append(ReplayWriter(open(viewer[1], 'w')))
        elif viewer in ('tk', 'tk-no-sync'):
            if not zmq_publisher:
                zmq_publisher = ZMQPublisher(address='tcp://127.0.0.1')
                viewer_state['viewers'].append(zmq_publisher)
            if viewer == 'tk':
                viewer_state['controller'] = setup_controller()
            if viewer_state['controller']:
                proc = TkViewer(address=zmq_publisher.socket_addr, controller=viewer_state['controller'].socket_addr,
                                stop_after=options.get('stop_at'),
                                geometry=options.get('geometry'),
                                delay=options.get('delay'),
                                fullscreen=options.get('fullscreen'))
            else:
                proc = TkViewer(address=zmq_publisher.socket_addr, controller=None,
                                stop_after=options.get('stop_at'),
                                geometry=options.get('geometry'),
                                delay=options.get('delay'),
                                fullscreen=options.get('fullscreen'))

        else:
            raise ValueError(f"Unknown viewer {viewer}.")

    # Add the result printer as the final viewer, if print_result has been given
    if print_result:
        viewer_state['viewers'].append(ResultPrinter())

    return viewer_state


def setup_game(team_specs, *, layout_dict, max_rounds=300, layout_name="", seed=None,
               allow_camping=False, error_limit=5, timeout_length=3,
               viewers=None, viewer_options=None, store_output=False,
               team_names=(None, None), team_infos=(None, None),
               allow_exceptions=False, print_result=True):
    """ Generates a game state for the given teams and layout with otherwise default values. """

    # check that two teams have been given
    if not len(team_specs) == 2:
        raise ValueError("Two teams must be given.")

    # check that the given bot positions are all valid
    # and None of them are None (`parse_layout` returns a None position
    # when fewer than 4 bots are defined)
    if not len(layout_dict['bots']) == 4 or None in layout_dict['bots']:
        raise ValueError("Number of bots in layout must be 4.")

    width, height = layout.wall_dimensions(layout_dict['walls'])
    if not (width, height) == layout_dict["shape"]:
        raise ValueError(f"layout_dict['walls'] does not match layout_dict['shape'].")

    for idx, pos in enumerate(layout_dict['bots']):
        if pos in layout_dict['walls']:
            raise ValueError(f"Bot {idx} must not be on wall: {pos}.")
        try:
            if len(pos) != 2:
                raise ValueError(f"Bot {idx} must be a valid position: got {pos} instead.")
        except TypeError:
            raise ValueError(f"Bot {idx} must be a valid position: got {pos} instead.")
        if not (0, 0) <= pos < (width, height):
            raise ValueError(f"Bot {idx} is not inside the layout: {pos}.")

    food = split_food(width, layout_dict['food'])
    max_food_age = math.inf if allow_camping else MAX_FOOD_AGE

    # warn if one of the food lists is already empty
    side_no_food = [idx for idx, f in enumerate(food) if len(f) == 0]
    if side_no_food:
        warn(f"Layout has no food for team {side_no_food}.", NoFoodWarning)

    viewer_state = setup_viewers(viewers, options=viewer_options, print_result=print_result)

    # Initialize the game state.

    game_state = dict(
        ### The layout attributes
        #: Walls. Set of (int, int)
        walls=set(layout_dict['walls']),

        #: Shape of the maze. (int, int)
        shape=layout_dict['shape'],

        #: Food per team. List of sets of (int, int)
        food=food,

        #: Food ages per team. Dict of (int, int) to int
        food_age=[{}, {}],

        ### Round/turn information
        #: Current bot, int, None
        turn=None,

        #: Current round, int, None
        round=None,

        #: Is the game finished? bool
        gameover=False,

        #: Who won? int, None
        whowins=None,

        ### Bot/team status
        #: Positions of all bots. List of (int, int)
        bots=layout_dict['bots'][:],

        #: Score of the teams. List of int
        score=[0] * 2,

        #: Fatal errors
        fatal_errors=[[], []],

        #: Errors
        errors=[{}, {}],

        ### Configuration
        #: Maximum number of rounds, int
        max_rounds=max_rounds,

        #: Time till timeout, int
        timeout=3,

        #: Noise radius, int
        noise_radius=NOISE_RADIUS,

        #: Sight distance, int
        sight_distance=SIGHT_DISTANCE,

        #: Max food age
        max_food_age=max_food_age,

        #: Shadow distance, int
        shadow_distance=SHADOW_DISTANCE,

        ### Informative
        #: Name of the layout, str
        layout_name=layout_name,

        #: Name of the teams. Tuple of str
        team_names=team_names,

        #: Additional team info. Tuple of str|None
        team_infos=team_infos,

        #: Time each team needed, list of float
        team_time=[0, 0],

        # List of bot deaths, which counts the number of deaths per bot
        # In other words, deaths[bot_idx] is the number of times the bot
        # bot_idx has been killed until now.
        deaths = [0]*4,

        # List of bot kills, which counts the number of kills per bot
        # In other words, kills[bot_idx] is the number of times the bot
        # bot_idx has killed another bot until now.
        kills = [0]*4,

        # List of boolean flags weather bot has been eaten since its last move
        bot_was_killed = [False]*4,

        # The noisy positions that the bot in `turn` has currently been shown.
        # None, if not noisy
        noisy_positions = [None] * 4,

        #: The moves that the bots returned. Keeps only the recent one at the respective bot’s index.
        requested_moves=[None] * 4,

        #: Messages the bots say. Keeps only the recent one at the respective bot’s index.
        say=[""] * 4,

        ### Internal
        #: Internal team representation
        teams=[None] * 2,

        #: Random number generator
        rnd=Random(seed),

        #: Timeout length, int, None
        timeout_length=timeout_length,

        #: Error limit. A team loses when the limit is reached, int
        error_limit=error_limit,

        #: Viewers, list
        viewers=viewer_state['viewers'],

        #: Controller
        controller=viewer_state['controller']
    )

    # Wait until the controller tells us that it is ready
    # We then can send the initial maze
    # This call *blocks* until the controller replies
    if controller_exit(game_state, await_action='set_initial'):
        return game_state

    # Send maze before team creation.
    # This gives a more fluent UI as it does not have to wait for the clients
    # to answer to the server.
    update_viewers(game_state)

    team_state = setup_teams(team_specs, game_state, store_output=store_output, allow_exceptions=allow_exceptions)
    game_state.update(team_state)

    # Check if one of the teams has already generate a fatal error
    # or if the game has finished (might happen if we set it up with max_rounds=0).
    game_state.update(check_gameover(game_state, detect_final_move=True))

    # Send updated game state with team names to the viewers
    update_viewers(game_state)

    # exit remote teams in case we are game over
    check_exit_remote_teams(game_state)

    return game_state


def setup_teams(team_specs, game_state, store_output=False, allow_exceptions=False):
    """ Creates the teams according to the `teams`. """

    # we start with a dummy zmq_context
    # make_team will generate and return a new zmq_context,
    # if it is needed for a remote team
    zmq_context = None

    teams = []
    # First, create all teams
    # If a team is a RemoteTeam, this will start a subprocess
    for idx, team_spec in enumerate(team_specs):
        team, zmq_context = make_team(team_spec, idx=idx, zmq_context=zmq_context, store_output=store_output, team_name=game_state['team_names'][idx])
        teams.append(team)

    # Send the initial state to the teams and await the team name (if the teams are local, the name can be get from the game_state directly
    team_names = []
    for idx, team in enumerate(teams):
        try:
            team_name = team.set_initial(idx, prepare_bot_state(game_state, idx))
        except (FatalException, PlayerTimeout) as e:
            # TODO: Not sure if PlayerTimeout should let the other payer win.
            # It could simply be a network problem.
            if allow_exceptions: raise
            exception_event = {
                'type': e.__class__.__name__,
                'description': str(e),
                'turn': idx,
                'round': None,
            }
            game_state['fatal_errors'][idx].append(exception_event)
            if len(e.args) > 1:
                game_print(idx, f"{type(e).__name__} ({e.args[0]}): {e.args[1]}")
                team_name = f"%%%{e.args[0]}%%%"
            else:
                game_print(idx, f"{type(e).__name__}: {e}")
                team_name = "%%%error%%%"
        team_names.append(team_name)

    team_state = {
        'teams': teams,
        'team_names': team_names
    }
    return team_state


def request_new_position(game_state):
    team = game_state['turn'] % 2
    bot_turn = game_state['turn'] // 2
    move_fun = game_state['teams'][team]

    bot_state = prepare_bot_state(game_state)


    start_time = time.monotonic()

    new_position = move_fun.get_move(bot_state)

    duration = time.monotonic() - start_time
    # update the team_time
    game_state['team_time'][team] += duration

    return new_position


def prepare_bot_state(game_state, idx=None):
    """ Prepares the bot’s game state for the current bot.

    """

    bot_initialization = game_state.get('turn') is None and idx is not None
    bot_finalization = game_state.get('turn') is not None and idx is not None

    if bot_initialization:
        # We assume that we are in get_initial phase
        turn = idx
        bot_turn = None
        seed = game_state['rnd'].randint(0, sys.maxsize)
    elif bot_finalization:
        # Called for remote players in _exit
        turn = idx
        bot_turn = None
        seed = None
    else:
        turn = game_state['turn']
        bot_turn = game_state['turn'] // 2
        seed = None

    bot_position = game_state['bots'][turn]
    own_team = turn % 2
    enemy_team = 1 - own_team
    enemy_positions = game_state['bots'][enemy_team::2]
    noised_positions = noiser(walls=game_state['walls'],
                              shape=game_state['shape'],
                              bot_position=bot_position,
                              enemy_positions=enemy_positions,
                              noise_radius=game_state['noise_radius'],
                              sight_distance=game_state['sight_distance'],
                              rnd=game_state['rnd'])


    # Update noisy_positions in the game_state
    # reset positions
    game_state['noisy_positions'] = [None] * 4
    noisy_or_none = [
        noisy_pos if is_noisy else None
            for is_noisy, noisy_pos in
            zip(noised_positions['is_noisy'], noised_positions['enemy_positions'])
    ]
    game_state['noisy_positions'][enemy_team::2] = noisy_or_none
    shaded_food = list(pos for pos, age in game_state['food_age'][own_team].items()
                       if age > 0)

    team_state = {
        'team_index': own_team,
        'bot_positions': game_state['bots'][own_team::2],
        'score': game_state['score'][own_team],
        'kills': game_state['kills'][own_team::2],
        'deaths': game_state['deaths'][own_team::2],
        'bot_was_killed': game_state['bot_was_killed'][own_team::2],
        'error_count': len(game_state['errors'][own_team]),
        'food': list(game_state['food'][own_team]),
        'shaded_food': shaded_food,
        'name': game_state['team_names'][own_team],
        'team_time': game_state['team_time'][own_team]
    }

    enemy_state = {
        'team_index': enemy_team,
        'bot_positions': noised_positions['enemy_positions'],
        'is_noisy': noised_positions['is_noisy'],
        'score': game_state['score'][enemy_team],
        'kills': game_state['kills'][enemy_team::2],
        'deaths': game_state['deaths'][enemy_team::2],
        'bot_was_killed': game_state['bot_was_killed'][enemy_team::2],
        'error_count': 0, # TODO. Could be left out for the enemy
        'food': list(game_state['food'][enemy_team]),
        'shaded_food': [],
        'name': game_state['team_names'][enemy_team],
        'team_time': game_state['team_time'][enemy_team]
    }

    bot_state = {
        'team': team_state,
        'enemy': enemy_state,
        'round': game_state['round'],
        'bot_turn': bot_turn,
        'timeout_length': game_state['timeout_length'],
        'max_rounds': game_state['max_rounds'],
    }

    if bot_initialization:
        bot_state.update({
            'walls': game_state['walls'], # only in initial round
            'shape': game_state['shape'], # only in initial round
            'seed': seed # only used in set_initial phase
        })

    return bot_state


def update_viewers(game_state):
    """ Sends the current game_state to the viewers. """
    viewer_state = prepare_viewer_state(game_state)
    for viewer in game_state['viewers']:
        viewer.show_state(viewer_state)


def prepare_viewer_state(game_state):
    """ Prepares a state that can be sent to the viewers by removing
    date that cannot be serialized (ie. sockets or keys that
    cannot be used in a json object).

    Furthermore, some redundant data is removed when it has
    already been sent at an earlier date.

    Returns
    -------
    viewer_state : dict
       a new state dict
    """
    viewer_state = {}
    viewer_state.update(game_state)

    # Flatten food and food_age
    viewer_state['food'] = list((viewer_state['food'][0] | viewer_state['food'][1]))
    # We must transform the food age dict to a list or we cannot serialise it
    viewer_state['food_age'] = [item for team_food_age in viewer_state['food_age']
                                          for item in team_food_age.items()]

    # game_state["errors"] has a tuple as a dict key
    # that cannot be serialized in json.
    # To fix this problem, we only send the current error
    # and add another attribute "num_errors"
    # to the final dict.

    # the key for the current round, turn
    round_turn = (game_state["round"], game_state["turn"])
    viewer_state["errors"] = [
        # retrieve the current error or None
        team_errors.get(round_turn)
        for team_errors in game_state["errors"]
    ]

    # add the number of errors
    viewer_state["num_errors"] = [
        len(team_errors)
        for team_errors in game_state["errors"]
    ]

    # remove unserializable values
    del viewer_state['teams']
    del viewer_state['rnd']
    del viewer_state['viewers']
    del viewer_state['controller']

    return viewer_state


def play_turn(game_state, allow_exceptions=False):
    """ Plays the next turn of the game.

    This function increases the round and turn counters, requests a move
    and returns a new game_state.

    Raises
    ------
    ValueError
        If gamestate['gameover'] is True
    """
    # TODO: Return a copy of the game_state

    # if the game is already over, we return a value error
    if game_state['gameover']:
        raise ValueError("Game is already over!")

    # Now update the round counter
    game_state.update(next_round_turn(game_state))

    turn = game_state['turn']
    round = game_state['round']
    team = turn % 2

    # update food age and relocate expired food for the current team
    game_state.update(update_food_age(game_state, team, SHADOW_DISTANCE))
    game_state.update(relocate_expired_food(game_state, team, SHADOW_DISTANCE))

    # request a new move from the current team
    try:
        position_dict = request_new_position(game_state)
        if "error" in position_dict:
            error_type, error_string = position_dict['error']
            raise FatalException(f"Exception in client ({error_type}): {error_string}")
        try:
            position = tuple(position_dict['move'])
        except TypeError as e:
            raise NonFatalException(f"Type error {e}")

        if position_dict.get('say'):
            game_state['say'][game_state['turn']] = position_dict['say']
        else:
            game_state['say'][game_state['turn']] = ""
    except FatalException as e:
        if allow_exceptions: raise
        # FatalExceptions (such as PlayerDisconnect) should immediately
        # finish the game
        exception_event = {
            'type': e.__class__.__name__,
            'description': str(e),
            'turn': game_state['turn'],
            'round': game_state['round'],
        }
        game_state['fatal_errors'][team].append(exception_event)
        position = None
        game_print(turn, f"{type(e).__name__}: {e}")
    except NonFatalException as e:
        if allow_exceptions: raise
        # NonFatalExceptions (such as Timeouts and ValueErrors in the JSON handling)
        # are collected and added to team_errors
        exception_event = {
            'type': e.__class__.__name__,
            'description': str(e)
        }
        game_state['errors'][team][(round, turn)] = exception_event
        position = None
        game_print(turn, f"{type(e).__name__}: {e}")

    # If the returned move looks okay, we add it to the list of requested moves
    old_position = game_state['bots'][turn]
    game_state['requested_moves'][turn] = {
        'previous_position': old_position,
        'requested_position': position,
        'success': False # Success is set to true after apply_move
    }

    # Check if a team has exceeded their maximum number of errors
    # (we do not want to apply the move in this case)
    # Note: Since we already updated the move counter, we do not check anymore,
    # if the game has exceeded its rounds.
    game_state.update(check_gameover(game_state))

    if not game_state['gameover']:
        # ok. we can apply the move for this team
        # try to execute the move and return the new state
        game_state = apply_move(game_state, position)

        # If there was no error, we claim a success in requested_moves
        if (round, turn) not in game_state["errors"][team]:
            game_state['requested_moves'][turn]['success'] = True

        # Check again, if we had errors or if this was the last move of the game (final round or food eaten)
        game_state.update(check_gameover(game_state, detect_final_move=True))

    # Send updated game state with team names to the viewers
    update_viewers(game_state)

    # exit remote teams in case we are game over
    check_exit_remote_teams(game_state)

    return game_state


def apply_move(gamestate, bot_position):
    """Plays a single step of a bot by applying the game rules to the game state. The rules are:
    - if the playing team has an error count of >4 or a fatal error they lose
    - a legal step must not be on a wall, else the error count is increased by 1 and a random move is chosen for the bot
    - if a bot lands on an enemy food pellet, it eats it. It cannot eat its own teams’ food
    - if a bot lands on an enemy bot in its own homezone, it kills the enemy
    - if a bot lands on an enemy bot in its enemy’s homezone, it dies
    - when a bot dies, it reappears in its own homezone at the initial position
    - a game ends when max_rounds is exceeded

    Parameters
    ----------
    gamestate : dict
        state of the game before current turn
    turn : int
        index of the current bot. 0, 1, 2, or 3.
    bot_position : tuple
        new coordinates (x, y) of the current bot.

    Returns
    -------
    dict
        state of the game after applying current turn

    """
    # TODO is a timeout counted as an error?
    # define local variables
    bots = gamestate["bots"]
    turn = gamestate["turn"]
    team = turn % 2
    enemy_idx = (1, 3) if team == 0 else (0, 2)
    gameover = gamestate["gameover"]
    score = gamestate["score"]
    food = gamestate["food"]
    walls = gamestate["walls"]
    shape = gamestate["shape"]
    food = gamestate["food"]
    n_round = gamestate["round"]
    kills = gamestate["kills"]
    deaths = gamestate["deaths"]
    bot_was_killed = gamestate["bot_was_killed"]
    fatal_error = True if gamestate["fatal_errors"][team] else False
    #TODO how are fatal errors passed to us? dict with same structure as regular errors?
    #TODO do we need to communicate that fatal error was the reason for game over in any other way?


    # reset our own bot_was_killed flag
    bot_was_killed[turn] = False

    # previous errors
    team_errors = gamestate["errors"][team]

    # the allowed moves for the current bot
    legal_positions = get_legal_positions(walls, shape, gamestate["bots"][gamestate["turn"]])

    # unless we have already made an error, check if we made a legal move
    if not (n_round, turn) in team_errors:
        if bot_position not in legal_positions:
            error_dict = {
                "reason": 'illegal move',
                "bot_position": bot_position
                }
            # add the error to the team’s errors
            game_print(turn, f"Illegal position. {bot_position} not in legal positions: {sorted(legal_positions)}.")
            team_errors[(n_round, turn)] = error_dict

    # only execute move if errors not exceeded
    gamestate.update(check_gameover(gamestate))
    if gamestate['gameover']:
        return gamestate

    # Now check if we must make a random move
    if (n_round, turn) in team_errors:
        # There was an error for this round and turn
        # but the game is not over.
        # We execute a random move
        bot_position = gamestate['rnd'].choice(legal_positions)
        game_print(turn, f"Setting a legal position at random: {bot_position}")

    # take step
    bots[turn] = bot_position
    _logger.info(f"Bot {turn} moves to {bot_position}.")
    # then apply rules
    # is bot in home or enemy territory
    boundary = gamestate['shape'][0] / 2
    if team == 0:
        bot_in_homezone = bot_position[0] < boundary
    elif team == 1:
        bot_in_homezone = bot_position[0] >= boundary
    # update food list
    if not bot_in_homezone:
        if bot_position in food[1 - team]:
            _logger.info(f"Bot {turn} eats food at {bot_position}.")
            food[1 - team].remove(bot_position)
            # This is modifying the old game state
            score[team] = score[team] + 1
    # check if we killed someone
    if bot_in_homezone:
        killed_enemies = [idx for idx in enemy_idx if bot_position == bots[idx]]
        for enemy_idx in killed_enemies:
            _logger.info(f"Bot {turn} eats enemy bot {enemy_idx} at {bot_position}.")
            score[team] = score[team] + KILL_POINTS
            init_positions = initial_positions(walls, shape)
            bots[enemy_idx] = init_positions[enemy_idx]
            kills[turn] += 1
            deaths[enemy_idx] += 1
            bot_was_killed[enemy_idx] = True
            _logger.info(f"Bot {enemy_idx} reappears at {bots[enemy_idx]}.")
    else:
        # check if we have been eaten
        enemies_on_target = [idx for idx in enemy_idx if bots[idx] == bot_position]
        if len(enemies_on_target) > 0:
            _logger.info(f"Bot {turn} was eaten by bots {enemies_on_target} at {bot_position}.")
            score[1 - team] = score[1 - team] + KILL_POINTS
            init_positions = initial_positions(walls, shape)
            bots[turn] = init_positions[turn]
            deaths[turn] += 1
            kills[enemies_on_target[0]] += 1
            bot_was_killed[turn] = True
            _logger.info(f"Bot {turn} reappears at {bots[turn]}.")

    errors = gamestate["errors"]
    errors[team] = team_errors
    gamestate_new = {
        "food": food,
        "bots": bots,
        "score": score,
        "deaths": deaths,
        "kills": kills,
        "bot_was_killed": bot_was_killed,
        "errors": errors,
        }

    gamestate.update(gamestate_new)
    return gamestate


def next_round_turn(game_state):
    """ Takes the round and turn from the game state dict and returns
    the round and turn of the next step in a dict.

    Returns
    -------
    dict { 'round' , 'turn' }
        The updated round and turn

    Raises
    ------
    ValueError
        If gamestate['gameover'] is True
    """

    if game_state['gameover']:
        raise ValueError("Game is already over")
    turn = game_state['turn']
    round = game_state['round']

    if turn is None and round is None:
        turn = 0
        round = 1
    else:
        # if one of turn or round is None bot not both, it is illegal.
        # TODO: fail with a better error message
        turn = turn + 1
        if turn >= 4:
            turn = turn % 4
            round = round + 1

    return {
        'round': round,
        'turn': turn,
    }


def check_gameover(game_state, detect_final_move=False):
    """ Checks if this was the final moves or if the errors have exceeded the threshold.

    Returns
    -------
    dict { 'gameover' , 'whowins' }
        Flags if the game is over and who won it
    """

    # If any team has a fatal error, this team loses.
    # If both teams have a fatal error, it’s a draw.
    num_fatals = [len(f) for f in game_state['fatal_errors']]
    if num_fatals[0] == 0 and num_fatals[1] == 0:
        # no one has any fatal errors
        pass
    elif num_fatals[0] > 0 and num_fatals[1] > 0:
        # both teams have fatal errors: it is a draw
        return { 'whowins' : 2, 'gameover' : True}
    else:
        # some one has fatal errors
        for team in (0, 1):
            if num_fatals[team] > 0:
                return { 'whowins' : 1 - team, 'gameover' : True}

    # If any team has reached error_limit errors, this team loses.
    # If both teams have reached error_limit errors, it’s a draw.
    # If error_limit is 0, the game will go on without checking.
    num_errors = [len(f) for f in game_state['errors']]
    if game_state['error_limit'] == 0:
        pass
    elif num_errors[0] < game_state['error_limit'] and num_errors[1] < game_state['error_limit']:
        # no one has reached the error limit
        pass
    elif num_errors[0] >= game_state['error_limit'] and num_errors[1] >= game_state['error_limit']:
        # both teams have reached or exeeded the error limit
        return { 'whowins' : 2, 'gameover' : True}
    else:
        # only one team has reached the error limit
        for team in (0, 1):
            if num_errors[team] >= game_state['error_limit']:
                return { 'whowins' : 1 - team, 'gameover' : True}

    if detect_final_move:
        # No team wins/loses because of errors?
        # Good. Now check if the game finishes because the food is gone
        # or because we are in the final turn of the last round.

        # will we overshoot the max rounds with the next step?
        next_step = next_round_turn(game_state)
        next_round = next_step['round']

        # count how much food is left for each team
        food_left = [len(f) for f in game_state['food']]
        if next_round > game_state['max_rounds'] or any(f == 0 for f in food_left):
            if game_state['score'][0] > game_state['score'][1]:
                whowins = 0
            elif game_state['score'][0] < game_state['score'][1]:
                whowins = 1
            else:
                whowins = 2
            return { 'whowins' : whowins, 'gameover' : True}

    return { 'whowins' : None, 'gameover' : False}


def check_exit_remote_teams(game_state):
    """ If the we are gameover, we want the remote teams to shut down. """
    if game_state['gameover']:
        _logger.info("Gameover. Telling teams to exit.")
        for idx, team in enumerate(game_state['teams']):
            if len(game_state['fatal_errors'][idx]) > 0:
                _logger.info(f"Not sending exit to team {idx} which had a fatal error.")
                # We pretend we already send the exit message, otherwise
                # the team’s __del__ method will do it once more.
                team._sent_exit = True
                continue
            try:
                team_game_state = prepare_bot_state(game_state, idx=idx)
                team._exit(team_game_state)
            except AttributeError:
                pass


def split_food(width, food):
    team_food = [set(), set()]
    for pos in food:
        idx = pos[0] // (width // 2)
        team_food[idx].add(pos)
    return team_food


def game_print(turn, msg):
    allow_unicode = not _mswindows
    if turn % 2 == 0:
        pie = ('\033[94m' + 'ᗧ' + '\033[0m' + ' ') if allow_unicode else ''
        pie += f'blue team, bot {turn // 2}'
    elif turn % 2 == 1:
        pie = ('\033[91m' + 'ᗧ' + '\033[0m' + ' ') if allow_unicode else ''
        pie += f'red team, bot {turn // 2}'
    print(f'{pie}: {msg}')




import logging
import os
import random
import subprocess
import sys
import traceback
from io import StringIO
from pathlib import Path
from urllib.parse import urlparse

import zmq
import networkx as nx

from . import layout
from .exceptions import PlayerDisconnected, PlayerTimeout
from .layout import layout_as_str, BOT_I2N, wall_dimensions
from .network import ZMQClientError, ZMQConnection, ZMQReplyTimeout, ZMQUnreachablePeer, PELITA_PORT

_logger = logging.getLogger(__name__)


def _ensure_list_tuples(list):
    """ Ensures that an iterable is a list of position tuples. """
    return [tuple(item) for item in list]

def _ensure_tuple_tuples(set):
    """ Ensures that an iterable is a tuple of position tuples. """
    return tuple(sorted(tuple(item) for item in set))


def create_homezones(shape, walls):
    width, height = shape
    return [
        tuple((x, y) for x in range(0, width // 2)
                     for y in range(0, height) if (x, y) not in walls),
        tuple((x, y) for x in range(width // 2, width)
                     for y in range(0, height) if (x, y) not in walls)
    ]

def walls_to_graph(walls):
    """Return a networkx Graph object given the walls of a maze.

    Parameters
    ----------
    walls : set[(x0,y0), (x1,y1), ...]
         a set of wall coordinates

    Returns
    -------
    graph : networkx.Graph
         a networkx Graph representing the free squares in the maze and their
         connections. Note that 'free' in this context means that the corresponding
         square in the maze is not a wall (but can contain food or bots).

    Notes
    -----
    Nodes in the graph are (x,y) coordinates representing squares in the maze
    which are not walls.
    Edges in the graph are ((x1,y1), (x2,y2)) tuples of coordinates of two
    adjacent squares. Adjacent means that you can go from one square to one of
    its adjacent squares by making ore single step (up, down, left, or right).
    """
    graph = nx.Graph()
    width, height = wall_dimensions(walls)

    for x in range(width):
        for y in range(height):
            if (x, y) not in walls:
                # this is a free position, get its neighbors
                for delta_x, delta_y in ((1,0), (-1,0), (0,1), (0,-1)):
                    neighbor = (x + delta_x, y + delta_y)
                    # we don't need to check for getting neighbors out of the maze
                    # because our mazes are all surrounded by walls, i.e. our
                    # deltas will not put us out of the maze
                    if neighbor not in walls:
                        # this is a genuine neighbor, add an edge in the graph
                        graph.add_edge((x, y), neighbor)
    return graph



class Team:
    """
    Wraps a move function and forwards it the `set_initial`
    and `get_move` requests.

    The Team class holds the team’s state between turns, the team’s
    random number generator and the bot track (resets every time a bot
    is killed). This class is also caching bot attributes that do not
    change during the game. Currently cached attributes:
        - bot.walls
        - bot.shape
        - bot._initial_position
        - bot.homezone
        - bot.graph

    Parameters
    ----------
    team_move : function with (bot, state) -> position
        the team’s move function
    team_name :
        the name of the team (optional)
    """
    def __init__(self, team_move, *, team_name=""):
        self._team_move = team_move
        self.team_name = team_name

        #: Storage for the team state
        self._state = {}

        #: The team’s random number generator
        self._random = None

        #: The history of bot positions
        self._bot_track = [[], []]


    def set_initial(self, team_id, game_state):
        """ Sets the bot indices for the team and returns the team name.
        Currently, we do not call _set_initial on the user side.

        Parameters
        ----------
        team_id : int
            The id of the team
        game_state : dict
            The initial game state

        Returns
        -------
        Team name : string
            The name of the team

        """
        # Reset the team state
        self._state.clear()

        # Initialize the random number generator
        # with the seed that we received from game
        self._random = random.Random(game_state['seed'])

        # Reset the bot tracks
        self._bot_track = [[], []]

        # Store the walls, which are only transmitted once
        self._walls = _ensure_tuple_tuples(game_state['walls'])

        # Store the shape, which is only transmitted once
        self._shape = tuple(game_state['shape'])

        # Cache the initial positions so that we don’t have to calculate them at each step
        self._initial_positions = layout.initial_positions(self._walls, self._shape)

        # Cache the homezone so that we don’t have to create it at each step
        self._homezone = create_homezones(self._shape, self._walls)

        # Cache the graph representation of the maze -> stores a read-only view of the
        # graph, so that local modifications in the move function are not carried
        # over
        self._graph = walls_to_graph(self._walls).copy(as_view=True)

        return self.team_name

    def get_move(self, game_state):
        """ Requests a move from the Player who controls the Bot with id `bot_id`.

        This method returns a dict with a key `move` and a value specifying the direction
        in a tuple. Additionally, a key `say` can be added with a textual value.

        Parameters
        ----------
        game_state : dict
            The initial game state

        Returns
        -------
        move : dict
        """
        me = make_bots(walls=self._walls,
                       shape=self._shape,
                       initial_positions=self._initial_positions,
                       homezone=self._homezone,
                       team=game_state['team'],
                       enemy=game_state['enemy'],
                       round=game_state['round'],
                       bot_turn=game_state['bot_turn'],
                       rng=self._random,
                       graph=self._graph)

        team = me._team

        for idx, mybot in enumerate(team):
            # If a bot has been killed, we reset its bot track
            if mybot.was_killed:
                self._bot_track[idx] = []

        # Add our track
        if len(self._bot_track[me._bot_turn]) == 0:
            self._bot_track[me._bot_turn] = [me.position]

        for idx, mybot in enumerate(team):
            # If the track of any bot is empty,
            # add its current position
            if me._bot_turn != idx:
                self._bot_track[idx].append(mybot.position)

            mybot.track = self._bot_track[idx][:]

        try:
            # request a move from the current bot
            move = self._team_move(team[me._bot_turn], self._state)

            # check that the returned value is a position tuple
            try:
                if len(move) != 2:
                    raise ValueError(f"Function move did not return a valid position: got {move} instead.")
            except TypeError:
                # Convert to ValueError
                raise ValueError(f"Function move did not return a valid position: got {move} instead.") from None
        except Exception as e:
            # Our client had an exception. We print a traceback and
            # return the type of the exception to the server.
            # If this is a remote player, then this will be detected in pelita_player
            # and pelita_player will close the connection automatically.
            traceback.print_exc()
            return {
                "error": (type(e).__name__, str(e)),
            }

        return {
            "move": move,
            "say": me._say
        }

    def _exit(self, game_state=None):
        """ Dummy function. Only needed for `RemoteTeam`. """
        pass

    def __repr__(self):
        return f'Team({self._team_move!r}, {self.team_name!r})'


class RemoteTeam:
    """ Start a child process with the given `team_spec` and handle
    communication with it through a zmq.PAIR connection.

    It also does some basic checks for correct return values and tries to
    terminate the child process once it is not needed anymore.

    Parameters
    ----------
    team_spec
        The string to pass as a command line argument to pelita_player
        or the address of a remote player
    team_name
        Overrides the team name
    zmq_context
        A zmq_context (if None, a new one will be created)
    idx
        The team index (currently only used to specify the team’s color)
    store_output
        If store_output is a string it will be interpreted as a path to a
        directory where to store stdout and stderr for the client processes.
        It helps in debugging issues with the clients.
        In the special case of store_output==subprocess.DEVNULL, stdout of
        the remote clients will be suppressed.
    """
    def __init__(self, team_spec, *, team_name=None, zmq_context=None, idx=None, store_output=False):
        if zmq_context is None:
            zmq_context = zmq.Context()

        self._team_spec = team_spec
        self._team_name = team_name

        #: Default timeout for a request, unless specified in the game_state
        self._request_timeout = 3

        if team_spec.startswith('pelita://'):
            # We connect to a remote player that is listening
            # on the given team_spec address.
            # We create a new DEALER socket and send a single
            # REQUEST message to the remote address.
            # The remote player will then create a new instance
            # of a player and forward all of our zmq traffic
            # to that player.

            # given a url pelita://hostname:port/path we extract hostname and port and
            # convert it to tcp://hostname:port that we use for the zmq connection
            parsed_url = urlparse(team_spec)
            if parsed_url.port:
                port = parsed_url.port
            else:
                port = PELITA_PORT
            send_addr = f"tcp://{parsed_url.hostname}:{port}"
            address = "tcp://*"
            self.bound_to_address = address

            socket = zmq_context.socket(zmq.DEALER)
            socket.setsockopt(zmq.LINGER, 0)
            socket.connect(send_addr)
            _logger.info("Connecting zmq.DEALER to remote player at {}.".format(send_addr))

            socket.send_json({"REQUEST": team_spec})
            WAIT_TIMEOUT = 5000
            incoming = socket.poll(timeout=WAIT_TIMEOUT)
            if incoming == zmq.POLLIN:
                ok = socket.recv()
            else:
                # Server did not respond
                raise PlayerTimeout()
            self.proc = None

        else:
            # We bind to a local tcp port with a zmq PAIR socket
            # and start a new subprocess of pelita_player.py
            # that includes the address of that socket and the
            # team_spec as command line arguments.
            # The subprocess will then connect to this address
            # and load the team.

            socket = zmq_context.socket(zmq.PAIR)
            port = socket.bind_to_random_port('tcp://*')
            self.bound_to_address = f"tcp://localhost:{port}"
            if idx == 0:
                color='blue'
            elif idx == 1:
                color='red'
            else:
                color=''
            self.proc = self._call_pelita_player(team_spec, self.bound_to_address,
                                                 color=color, store_output=store_output)

        self.zmqconnection = ZMQConnection(socket)

    def _call_pelita_player(self, team_spec, address, color='', store_output=False):
        """ Starts another process with the same Python executable,
        the same start script (pelitagame) and runs `team_spec`
        as a standalone client on URL `addr`.
        """
        player = 'pelita.scripts.pelita_player'
        external_call = [sys.executable,
                         '-m',
                         player,
                         'remote-game',
                         team_spec,
                         address]

        _logger.debug("Executing: %r", external_call)
        if store_output == subprocess.DEVNULL:
            return (subprocess.Popen(external_call, stdout=store_output), None, None)
        elif store_output:
            store_path = Path(store_output)
            stdout = (store_path / f"{color or team_spec}.out").open('w')
            stderr = (store_path / f"{color or team_spec}.err").open('w')

            # We must run in unbuffered mode to enforce flushing of stdout/stderr,
            # otherwise we may lose some of what is printed
            proc = subprocess.Popen(external_call, stdout=stdout, stderr=stderr,
                                    env=dict(os.environ, PYTHONUNBUFFERED='x'))
            return (proc, stdout, stderr)
        else:
            return (subprocess.Popen(external_call), None, None)

    @property
    def team_name(self):
        if self._team_name is not None:
            return self._team_name

        try:
            msg_id = self.zmqconnection.send("team_name", {})
            team_name = self.zmqconnection.recv_timeout(msg_id, self._request_timeout)
            if team_name:
                self._team_name = team_name
            return team_name
        except ZMQReplyTimeout:
            _logger.info("Detected a timeout, returning a string nonetheless.")
            return "%error%"
        except ZMQUnreachablePeer:
            _logger.info("Detected a DeadConnection, returning a string nonetheless.")
            return "%error%"

    def set_initial(self, team_id, game_state):
        timeout_length = game_state['timeout_length']
        try:
            msg_id = self.zmqconnection.send("set_initial", {"team_id": team_id,
                                                    "game_state": game_state})
            team_name = self.zmqconnection.recv_timeout(msg_id, timeout_length)
            if team_name:
                self._team_name = team_name
            return team_name
        except ZMQReplyTimeout:
            # answer did not arrive in time
            raise PlayerTimeout()
        except ZMQUnreachablePeer:
            _logger.info("Could not properly send the message. Maybe just a slow client. Ignoring in set_initial.")
        except ZMQClientError as e:
            error_message = e.message
            error_type = e.error_type
            _logger.warning(f"Client connection failed ({error_type}): {error_message}")
            raise PlayerDisconnected(*e.args) from None

    def get_move(self, game_state):
        timeout_length = game_state['timeout_length']
        try:
            msg_id = self.zmqconnection.send("get_move", {"game_state": game_state})
            reply = self.zmqconnection.recv_timeout(msg_id, timeout_length)
            # make sure it is a dict
            reply = dict(reply)
            if "error" in reply:
                return reply
            # make sure that the move is a tuple
            reply["move"] = tuple(reply.get("move"))
            return reply
        except ZMQReplyTimeout:
            # answer did not arrive in time
            raise PlayerTimeout()
        except TypeError:
            # if we could not convert into a tuple or dict (e.g. bad reply)
            return None
        except ZMQUnreachablePeer:
            # if the remote connection is closed
            raise PlayerDisconnected()
        except ZMQClientError:
            raise

    def _exit(self, game_state=None):
        # We only want to exit once.
        if getattr(self, '_sent_exit', False):
            return

        if game_state:
            payload = {'game_state': game_state}
        else:
            payload = {}

        try:
            # TODO: make zmqconnection stateful. set flag when already disconnected
            # For now, we simply check the state of the socket so that we do not send
            # over an already closed socket.
            if self.zmqconnection.socket.closed:
                return
            # TODO: Include final state with exit message
            self.zmqconnection.send("exit", payload, timeout=1)
            self._sent_exit = True
        except ZMQUnreachablePeer:
            _logger.info("Remote Player %r is already dead during exit. Ignoring.", self)

    def __del__(self):
        try:
            self._exit()
            if self.proc:
                self.proc[0].terminate()
        except AttributeError:
            # in case we exit before self.proc or self.zmqconnection have been set
            pass

    def __repr__(self):
        team_name = f" ({self._team_name})" if self._team_name else ""
        return f"RemoteTeam<{self._team_spec}{team_name} on {self.bound_to_address}>"


def make_team(team_spec, team_name=None, zmq_context=None, idx=None, store_output=False):
    """ Creates a Team object for the given team_spec.

    If no zmq_context is passed for a remote team, then a new context
    will be automatically created and returned. Otherwise, the same
    zmq_context (or None) will be returned.

    Parameters
    ----------
    team_spec : callable or str
        A move function or a team_spec that is passed on to pelita_player

    team_name : str, optional
        Optional team name for a local team

    zmq_context : zmq context, optional
        ZMQ context to avoid having to create a new context for every team

    Returns
    -------
    team_player, zmq_context : tuple
        The team class to interact with
        The new ZMQ context

    """
    if callable(team_spec):
        _logger.info("Making a local team for %s", team_spec)
        # wrap the move function in a Team
        if team_name is None:
            team_name = f'local-team ({team_spec.__name__})'
        team_player = Team(team_spec, team_name=team_name)
    elif isinstance(team_spec, str):
        _logger.info("Making a remote team for %s", team_spec)
        # set up the zmq connections and build a RemoteTeam
        if not zmq_context:
            zmq_context = zmq.Context()
        team_player = RemoteTeam(team_spec=team_spec, zmq_context=zmq_context, idx=idx, store_output=store_output)
    else:
        raise TypeError(f"Not possible to create team from {team_spec} (wrong type).")

    return team_player, zmq_context


class Bot:
    def __init__(self, *, bot_index,
                          is_on_team,
                          position,
                          initial_position,
                          walls,
                          shape,
                          homezone,
                          food,
                          shaded_food,
                          score,
                          kills,
                          deaths,
                          was_killed,
                          random,
                          graph,
                          round,
                          bot_char,
                          is_blue,
                          team_name,
                          team_time,
                          error_count,
                          is_noisy,
                          bot_turn=None):
        self._bots = None
        self._say = None

        #: The previous positions of this bot including the current one.
        self.track = []

        #: Is this a friendly bot?
        self._is_on_team = is_on_team

        self.random = random
        self.position = tuple(position)
        self._initial_position = tuple(initial_position)
        self.walls = walls

        self.homezone = homezone
        self.food = food
        self.shaded_food = shaded_food
        self.shape = shape
        self.score  = score
        self.kills = kills
        self.deaths = deaths
        self.was_killed = was_killed
        self._bot_index  = bot_index
        self.round = round
        self.char = bot_char
        self.is_blue = is_blue
        self.team_name = team_name
        self.team_time = team_time
        self.error_count = error_count
        self.is_noisy = is_noisy
        self.graph = graph

        # The legal positions that the bot can reach from its current position,
        # including the current position.
        self.legal_positions = []

        for direction in [(0, 0), (-1, 0), (1, 0), (0, 1), (0, -1)]:
            new_pos = (position[0] + direction[0],
                       position[1] + direction[1])
            if not new_pos in self.walls:
                self.legal_positions.append(new_pos)

        # Attributes for Bot
        if self._is_on_team:
            assert bot_turn is not None
            self._bot_turn = bot_turn

    @property
    def _team(self):
       """ Both of our bots.
       """
       if self._is_on_team:
           return self._bots['team']
       else:
           return self._bots['enemy']

    @property
    def turn(self):
        """ The turn of our bot. """
        return self._bot_index % 2

    @property
    def other(self):
        """ The other bot in our team. """
        return self._team[1 - self.turn]

    @property
    def enemy(self):
        """ The list of enemy bots
        """
        if not self._is_on_team:
            return self._bots['team']
        else:
            return self._bots['enemy']

    def say(self, text):
        """ Print some text in the graphical interface. """
        self._say = text

    # def get_direction(self, position):
        # """ Return the direction needed to get to the given position.

        # Raises
        # ======
        # ValueError
            # If the position cannot be reached by a legal move
        # """
        # direction = (position[0] - self.position[0], position[1] - self.position[1])
        # if direction not in self.legal_directions:
            # raise ValueError("Cannot reach position %s (would have been: %s)." % (position, direction))
        # return direction

    # def get_position(self, direction):
        # """ Return the position reached with the given direction

        # Raises
        # ======
        # ValueError
            # If the direction is not legal.
        # """
        # if direction not in self.legal_directions:
            # raise ValueError(f"Direction {direction} is not legal.")
        # position = (direction[0] + self.position[0], direction[1] + self.position[1])
        # return position


    def _repr_html_(self):
        """ Jupyter-friendly representation. """
        bot = self
        width, height = bot.shape

        with StringIO() as out:
            out.write("<table>")
            for y in range(height):
                out.write("<tr>")
                for x in range(width):
                    if (x, y) in bot.walls:
                        bg = 'style="background-color: {}"'.format(
                            "rgb(94, 158, 217)" if x < width // 2 else
                            "rgb(235, 90, 90)")
                    else:
                        bg = ""
                    out.write("<td %s>" % bg)
                    if (x, y) in bot.walls: out.write("#")
                    if (x, y) in bot.food: out.write('<span style="color: rgb(247, 150, 213)">●</span>')
                    if (x, y) in bot.enemy[0].food: out.write('<span style="color: rgb(247, 150, 213)">●</span>')
                    for idx in range(2):
                        if bot._team[idx].position == (x, y):
                            if idx == bot.turn:
                                out.write('<b>' + str(idx) + '</b>')
                            else:
                                out.write(str(idx))
                    for idx in range(2):
                        if bot.enemy[idx].position == (x, y):
                            out.write(str(idx))

                    out.write("</td>")
                out.write("</tr>")
            out.write("</table>")
            return out.getvalue()

    def __str__(self):
        bot = self

        if bot.is_blue:
            blue = bot if not bot.turn else bot.other
            red = bot.enemy[0]
        else:
            blue = bot.enemy[0]
            red = bot if not bot.turn else bot.other

        bot_positions = [blue.position, red.position, blue.other.position, red.other.position]
        bot_noise = [blue.is_noisy, red.is_noisy, blue.other.is_noisy, red.other.is_noisy]

        header = ("{blue}{you_blue} vs {red}{you_red}.\n" +
                  "Playing on {col} side. Current turn: {turn}. "+
                  "Bot: {bot_char}. Round: {round}, score: {blue_score}:{red_score}. " +
                  "timeouts: {blue_timeouts}:{red_timeouts}\n").format(
            blue=blue.team_name,
            red=red.team_name,
            turn=bot.turn,
            round=bot.round,
            bot_char=bot.char,
            blue_score=blue.score,
            red_score=red.score,
            col="blue" if bot.is_blue else "red",
            you_blue=" (you)" if bot.is_blue else "",
            you_red=" (you)" if not bot.is_blue else "",
            blue_timeouts=blue.error_count,
            red_timeouts=red.error_count,
        )

        footer = ("Bots: {bots}\nNoisy: {noise}\nFood: {food}\n").format(
                  bots={BOT_I2N[idx]:pos for idx, pos in enumerate(bot_positions)},
                  noise={BOT_I2N[idx]:state for idx, state in enumerate(bot_noise)},
                  food = bot.food + bot.enemy[0].food)

        with StringIO() as out:
            out.write(header)

            layout = layout_as_str(walls=bot.walls,
                                   food=bot.food + bot.enemy[0].food,
                                   bots=bot_positions,
                                   shape=bot.shape)

            out.write(str(layout))
            out.write(footer)
            return out.getvalue()


# def __init__(self, *, bot_index, position, initial_position, walls, homezone, food, is_noisy, score, random, round, is_blue):
def make_bots(*, walls, shape, initial_positions, homezone, team, enemy, round, bot_turn, rng, graph):
    bots = {}

    team_index = team['team_index']
    enemy_index = enemy['team_index']

    team_initial_positions = initial_positions[team_index::2]
    enemy_initial_positions = initial_positions[enemy_index::2]

    team_bots = []
    for idx, position in enumerate(team['bot_positions']):
        b = Bot(bot_index=idx,
            is_on_team=True,
            score=team['score'],
            deaths=team['deaths'][idx],
            kills=team['kills'][idx],
            was_killed=team['bot_was_killed'][idx],
            is_noisy=False,
            error_count=team['error_count'],
            food=_ensure_list_tuples(team['food']),
            shaded_food=_ensure_list_tuples(team['shaded_food']),
            walls=walls,
            shape=shape,
            round=round,
            bot_turn=bot_turn,
            bot_char=BOT_I2N[team_index + idx*2],
            random=rng,
            graph=graph,
            position=team['bot_positions'][idx],
            initial_position=team_initial_positions[idx],
            is_blue=team_index % 2 == 0,
            homezone=homezone[team_index],
            team_name=team['name'],
            team_time=team['team_time'])
        b._bots = bots
        team_bots.append(b)

    enemy_bots = []
    for idx, position in enumerate(enemy['bot_positions']):
        b = Bot(bot_index=idx,
            is_on_team=False,
            score=enemy['score'],
            kills=enemy['kills'][idx],
            deaths=enemy['deaths'][idx],
            was_killed=enemy['bot_was_killed'][idx],
            is_noisy=enemy['is_noisy'][idx],
            error_count=enemy['error_count'],
            food=_ensure_list_tuples(enemy['food']),
            shaded_food=[],
            walls=walls,
            shape=shape,
            round=round,
            bot_char = BOT_I2N[team_index + idx*2],
            random=rng,
            graph=graph,
            position=enemy['bot_positions'][idx],
            initial_position=enemy_initial_positions[idx],
            is_blue=enemy_index % 2 == 0,
            homezone=homezone[enemy_index],
            team_name=enemy['name'],
            team_time=enemy['team_time'])
        b._bots = bots
        enemy_bots.append(b)

    bots['team'] = team_bots
    bots['enemy'] = enemy_bots
    return team_bots[bot_turn]




"""Tests for Pelita game module"""
import pytest

from contextlib import contextmanager
import inspect
import itertools
import os
from pathlib import Path
import random
from textwrap import dedent
import sys

from pelita import game, layout
from pelita.exceptions import NoFoodWarning
from pelita.game import initial_positions, get_legal_positions, apply_move, run_game, setup_game, play_turn
from pelita.player import stepping_player, stopping_player

_mswindows = (sys.platform == "win32")

FIXTURE_DIR = Path(__file__).parent.resolve() / 'fixtures'

small_layout = """
################
#    . ##. . .y#
#  #       #.#x#
#. #.     .  ..#
#..  .     .# .#
#a#.#       #  #
#b. . .## .    #
################
"""

@contextmanager
def temp_wd(path):
    """ Temporarily change the working directory to path. """
    old = os.getcwd()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(old)


def test_too_few_registered_teams():
    test_layout_4 = (
    """ ##################
        #a#.  .  # .     #
        #b#####    #####x#
        #     . #  .  .#y#
        ################## """)
    team_1 = stopping_player
    with pytest.raises(ValueError):
        setup_game([team_1], layout_dict=layout.parse_layout(test_layout_4), max_rounds=300)


def test_too_many_registered_teams():
    test_layout_4 = (
    """ ##################
        #a#.  .  # .     #
        #b#####    #####x#
        #     . #  .  .#y#
        ################## """)
    team_1 = stopping_player
    with pytest.raises(ValueError):
        setup_game([team_1] * 3, layout_dict=layout.parse_layout(test_layout_4), max_rounds=300)


@pytest.mark.parametrize('layout_str', [
    """
    ######
    #a y #
    # b x#
    ######
    """,
    """
    ######
    #a y.#
    # b x#
    ######
    """])
def test_no_food(layout_str):
    with pytest.warns(NoFoodWarning):
        parsed = layout.parse_layout(layout_str)
        setup_game([stopping_player, stopping_player], layout_dict=parsed, max_rounds=300)


def test_initial_positions_basic():
    """Checks basic example for initial positions"""
    simple_layout = """
    ########
    #a ##b #
    #x   y #
    ########
    """
    parsed = layout.parse_layout(simple_layout)
    out = initial_positions(parsed['walls'], parsed['shape'])
    exp = [(1, 1), (6, 2), (1, 2), (6, 1)]
    assert len(out) == 4
    assert out == exp


@pytest.mark.parametrize('simple_layout', [
    # We use these test layouts to check that our algorithm finds
    # the expected initial position. This is noted by the location
    # of the respective bots in the layout.
    """
    ########
    #a### y#
    #b    x#
    ########
    """,
    """
    ########
    ##### y#
    #ba   x#
    ########
    """,
    """
    ########
    #a###xy#
    #b    ##
    ########
    """,
    """
    ########
    #####x##
    ##ba  y#
    ########
    """,
    ])
def test_initial_positions(simple_layout):
    parsed = layout.parse_layout(simple_layout)
    i_pos = initial_positions(parsed['walls'], parsed['shape'])
    expected = parsed['bots']
    assert len(i_pos) == 4
    assert i_pos == expected


@pytest.mark.parametrize('layout_t', [layout.get_random_layout() for _ in range(30)])
def test_initial_positions_same_in_layout_random(layout_t):
    """Check initial positions are the same as what the layout says for 30 random layouts"""
    layout_name, layout_string = layout_t # get_random_layout returns a tuple of name and string
    parsed_l = layout.parse_layout(layout_string)
    exp = parsed_l["bots"]
    walls = parsed_l["walls"]
    shape = parsed_l["shape"]
    out = initial_positions(walls, shape)
    assert out == exp

@pytest.mark.parametrize('layout_name', layout.get_available_layouts())
def test_initial_positions_same_in_layout(layout_name):
    """Check initial positions are the same as what the layout says for all layouts"""
    l = layout.get_layout_by_name(layout_name=layout_name)
    parsed_l = layout.parse_layout(l)
    exp = parsed_l["bots"]
    walls = parsed_l["walls"]
    shape = parsed_l["shape"]
    out = initial_positions(walls, shape)
    assert out == exp

def test_get_legal_positions_basic():
    """Check that the output of legal moves contains all legal moves for one example layout"""
    parsed_l = layout.parse_layout(small_layout)
    legal_positions = get_legal_positions(parsed_l["walls"], parsed_l["shape"], parsed_l["bots"][0])
    exp = [(1, 4), (1, 6), (1, 5)]
    assert legal_positions == exp

@pytest.mark.parametrize('layout_t', [layout.get_random_layout() for _ in range(50)])
@pytest.mark.parametrize('bot_idx', (0, 1, 2, 3))
def test_get_legal_positions_random(layout_t, bot_idx):
    """Check that the output of legal moves returns only moves that are 1 field away and not inside a wall"""
    layout_name, layout_string = layout_t # get_random_layout returns a tuple of name and string
    parsed_l = layout.parse_layout(layout_string)
    bot = parsed_l["bots"][bot_idx]
    legal_positions = get_legal_positions(parsed_l["walls"], parsed_l["shape"], bot)
    for move in legal_positions:
        assert move not in parsed_l["walls"]
        assert  abs((move[0] - bot[0])+(move[1] - bot[1])) <= 1

@pytest.mark.parametrize('turn', (0, 1, 2, 3))
def test_play_turn_apply_error(turn):
    """check that quits when there are too many errors"""
    game_state = setup_random_basic_gamestate()
    error_dict = {
        "reason": 'illegal move',
        "bot_position": (1, 2)
    }
    game_state["turn"] = turn
    team = turn % 2
    game_state["errors"] = [{(r, t): error_dict for r in (1, 2) for t in (0, 1)},
                            {(r, t): error_dict for r in (1, 2) for t in (0, 1)}]
    # we pretend that two rounds have already been played
    # so that the error dictionaries are sane
    game_state["round"] = 3

    illegal_position = (0, 0) # should always be a wall
    game_state_new = apply_move(game_state, illegal_position)
    assert game_state_new["gameover"]
    assert len(game_state_new["errors"][team]) == 5
    assert game_state_new["whowins"] == int(not team)
    assert set(game_state_new["errors"][team][(3, turn)].keys()) == set(["reason", "bot_position"])

@pytest.mark.parametrize('turn', (0, 1, 2, 3))
def test_play_turn_fatal(turn):
    """Checks that game quite after fatal error"""
    game_state = setup_random_basic_gamestate()
    game_state["turn"] = turn
    team = turn % 2
    fatal_list = [{}, {}]
    fatal_list[team] = {"error":True}
    game_state["fatal_errors"] = fatal_list
    move = get_legal_positions(game_state["walls"], game_state["shape"], game_state["bots"][turn])
    game_state_new = apply_move(game_state, move[0])
    assert game_state_new["gameover"]
    assert game_state_new["whowins"] == int(not team)

@pytest.mark.parametrize('turn', (0, 1, 2, 3))
def test_play_turn_illegal_position(turn):
    """check that illegal moves are added to error dict and bot still takes move"""
    game_state = setup_random_basic_gamestate()
    game_state["turn"] = turn
    team = turn % 2
    illegal_position = (0, 0) # should always be a wall
    game_state_new = apply_move(game_state, illegal_position)
    assert len(game_state_new["errors"][team]) == 1
    assert game_state_new["errors"][team][(1, turn)].keys() == set(["reason", "bot_position"])
    assert game_state_new["bots"][turn] in get_legal_positions(game_state["walls"], game_state["shape"], game_state["bots"][turn])

@pytest.mark.parametrize('turn', (0, 1, 2, 3))
@pytest.mark.parametrize('which_food', (0, 1))
def test_play_turn_eating_enemy_food(turn, which_food):
    """Check that you eat enemy food but not your own"""
    ### 012345678901234567
    #0# ##################
    #1# #. ... .##.     y#
    #2# # # #  .  .### #x#
    #3# # # ##.   .      #
    #4# #      .   .## # #
    #5# #a# ###.  .  # # #
    #6# #2     .##. ... .#
    #7# ##################
    game_state = setup_specific_basic_gamestate(round=0, turn=turn)
    team = turn % 2
    prev_len_food = [len(team_food) for team_food in game_state["food"]]

    if which_food == 0:
        # Try to eat food on left side
        game_state["bots"][turn] = (6, 4)
        move = (7, 4)
    else:
        # Try to eat food on right side
        game_state["bots"][turn] = (11, 1)
        move = (10, 1)

    game_state_new = apply_move(game_state, move)

    if team == which_food:
        # No changes for either team
        assert game_state_new["score"][team] == 0
        assert game_state_new["score"][1 - team] == 0
        assert prev_len_food[team] == len(game_state_new["food"][team])
        assert prev_len_food[1 - team] == len(game_state_new["food"][1 - team])
    elif team != which_food:
        # Own team gains points, other team loses food
        assert game_state_new["score"][team] > 0
        assert game_state_new["score"][1 - team] == 0
        assert prev_len_food[team] == len(game_state_new["food"][team])
        assert prev_len_food[1 - team] > len(game_state_new["food"][1 - team])


@pytest.mark.parametrize('turn', (0, 1, 2, 3))
def test_play_turn_killing(turn):
    """Check that you can kill enemies but not yourself"""
    ### 012345678901234567
    #0# ##################
    #1# #. ... .##.     y#
    #2# # # #  .  .### #x#
    #3# # # ##.   .      #
    #4# #      .   .## # #
    #5# #a# ###.  .  # # #
    #6# #b     .##. ... .#
    #7# ##################
    game_state = setup_specific_basic_gamestate()
    team = turn % 2
    game_state["turn"] = turn
    enemy_idx = (1, 3) if team == 0 else(0, 2)
    (friend_idx,) = set([0, 1, 2, 3]) - set([*enemy_idx, turn])

    game_state_new = apply_move(game_state, game_state["bots"][friend_idx])
    # assert game_state_new["DEATHS"][team] == 5
    assert game_state_new["score"] == [0, 0]
    assert game_state_new["deaths"] == [0]*4

@pytest.mark.parametrize('setups', ((0, (1, 4)),
                                    (1, (16, 3)),
                                    (2, (2, 6)),
                                    (3, (15, 1))))
def test_play_turn_friendly_fire(setups):
    """Check that you can kill enemies but not yourself"""
    ### 012345678901234567
    #0# ##################
    #1# #. ... .##.     y#
    #2# # # #  .  .### #x#
    #3# # # ##.   .      #
    #4# #      .   .## # #
    #5# #a# ###.  .  # # #
    #6# #b     .##. ... .#
    #7# ##################
    game_state = setup_specific_basic_gamestate()
    turn = setups[0]
    enemy_pos = setups[1]
    team = turn % 2
    game_state["turn"] = turn
    enemy_idx = (1, 3) if team == 0 else (0, 2)
    game_state["bots"][enemy_idx[0]] = enemy_pos
    game_state_new = apply_move(game_state, enemy_pos)
    # assert game_state_new["DEATHS"][team] == 5
    assert game_state_new["score"][team] == 5


def test_multiple_enemies_killing():
    """ Check that you can kill multiple enemies at once. """

    l0 = """
    ########
    #  ..  #
    # bxa  #
    ########
    """

    l1 = """
    ########
    #  ..  #
    #  xay #
    ########
    """
    # dummy bots
    stopping = lambda bot, s: (bot.position, s)

    parsed_l0 = layout.parse_layout(l0, bots={'y':(3,2)})
    for bot in (0, 2):
        game_state = setup_game([stopping, stopping], layout_dict=parsed_l0)

        game_state['turn'] = bot
        # get position of bots x (and y)
        kill_position = game_state['bots'][1]
        assert kill_position == game_state['bots'][3]
        new_state = apply_move(game_state, kill_position)
        # team 0 scores twice
        assert new_state['score'] == [10, 0]
        # bots 1 and 3 are back to origin
        assert new_state['bots'][1::2] == [(6, 2), (6, 1)]

    parsed_l1 = layout.parse_layout(l1, bots={'b':(4,2)})
    for bot in (1, 3):
        game_state = setup_game([stopping, stopping], layout_dict=parsed_l1)

        game_state['turn'] = bot
        # get position of bots 0 (and 2)
        kill_position = game_state['bots'][0]
        assert kill_position == game_state['bots'][2]
        new_state = apply_move(game_state, kill_position)
        # team 1 scores twice
        assert new_state['score'] == [0, 10]
        # bots 0 and 2 are back to origin
        assert new_state['bots'][0::2] == [(1, 1), (1, 2)]


def test_suicide():
    """ Check that suicide works. """

    l0 = """
    ########
    #  ..  #
    #ybxa  #
    ########
    """

    l1 = """
    ########
    #  ..  #
    #  xayb#
    ########
    """
    # dummy bots
    stopping = lambda bot, s: (bot.position, s)

    parsed_l0 = layout.parse_layout(l0)
    for bot in (1, 3):
        game_state = setup_game([stopping, stopping], layout_dict=parsed_l0)

        game_state['turn'] = bot
        # get position of bot 2
        suicide_position = game_state['bots'][2]
        new_state = apply_move(game_state, suicide_position)
        # team 0 scores
        assert new_state['score'] == [5, 0]
#        # bots 1 and 3 are back to origin
        if bot == 1:
            assert new_state['bots'][1::2] == [(6, 2), (1, 2)]
        elif bot == 3:
            assert new_state['bots'][1::2] == [(3, 2), (6, 1)]

    parsed_l1 = layout.parse_layout(l1)
    for bot in (0, 2):
        game_state = setup_game([stopping, stopping], layout_dict=parsed_l1)

        game_state['turn'] = bot
        # get position of bot 3
        suicide_position = game_state['bots'][3]
        new_state = apply_move(game_state, suicide_position)
        # team 0 scores
        assert new_state['score'] == [0, 5]


def test_cascade_kill():
    cascade = [
    ("""
    ########
    #x ..ya#
    #     b#
    ########
    """, {}),

    ("""
    ########
    #a .. y#
    #     b#
    ########
    """, {'x':(1,1)}),

    ("""
    ########
    #a .. y#
    #     x#
    ########
    """, {'b':(6,2)}),

    ("""
    ########
    #a .. y#
    #b    x#
    ########
    """, {}),
    ]

    def move(bot, state):
        if not bot.is_blue and bot.turn == 1 and bot.round == 1:
            return (6, 1)
        return bot.position
    layouts = [layout.parse_layout(l, bots=b) for l,b in cascade]
    state = setup_game([move, move], max_rounds=5, layout_dict=layouts[0])
    assert state['bots'] == layouts[0]['bots']
    state = game.play_turn(state) # Bot 0 stands
    assert state['bots'] == layouts[0]['bots']
    state = game.play_turn(state) # Bot 1 stands
    state = game.play_turn(state) # Bot 2 stands
    state = game.play_turn(state) # Bot 3 moves, kills 0. Bot 0 and 1 are on same spot
    assert state['bots'] == layouts[1]['bots']
    state = game.play_turn(state) # Bot 0 stands, kills 1. Bot 1 and 2 are on same spot
    assert state['bots'] == layouts[2]['bots']
    state = game.play_turn(state) # Bot 1 stands, kills 2.
    assert state['bots'] == layouts[3]['bots']


def test_cascade_kill_2():
    """ Checks that killing occurs only for the bot whose turn it is
    or for any bot that this bot moves onto.
    If a bot respawns on an enemy, it will only be killed when it is its own
    or the enemy’s turn (and neither of them moves).
    """
    cascade = [
    ("""
    ########
    #ya.. b#
    #x     #
    ########
    """,{}),

    ("""
    ########
    #a .. b#
    #x     #
    ########
    """, {'y':(6,1)}),

    ("""
    ########
    #a .. y#
    #x     #
    ########
    """, {'b':(1,2)}),

    ("""
    ########
    #a .. y#
    #b    x#
    ########
    """, {}),
    ]
    def move(bot, state):
        if bot.is_blue and bot.turn == 0 and bot.round == 1:
            return (1, 1)
        return bot.position
    layouts = [layout.parse_layout(l, bots=b) for l,b in cascade]
    state = setup_game([move, move], max_rounds=5, layout_dict=layouts[0])
    assert state['bots'] == layouts[0]['bots']
    state = game.play_turn(state) # Bot 0 moves, kills 3. Bot 2 and 3 are on same spot
    assert state['bots'] == layouts[1]['bots']
    state = game.play_turn(state) # Bot 1 stands. Bot 2 and 3 are on same spot
    assert state['bots'] == layouts[1]['bots']
    state = game.play_turn(state) # Bot 2 stands, gets killed. Bot 1 and 2 are on same spot
    assert state['bots'] == layouts[2]['bots']
    state = game.play_turn(state) # Bot 3 stands. Bot 1 and 2 are on same spot
    assert state['bots'] == layouts[2]['bots']
    state = game.play_turn(state) # Bot 0 stands. Bot 1 and 2 are on same spot
    assert state['bots'] == layouts[2]['bots']
    state = game.play_turn(state) # Bot 1 stands, kills 2.
    assert state['bots'] == layouts[3]['bots']


def test_cascade_kill_rescue_1():
    """ Checks that killing occurs only for the bot whose turn it is
    or for any bot that this bot moves onto.
    If a bot respawns on an enemy, it will only be killed when it is its own
    or the enemy’s turn (and neither of them moves).
    If bot moves before it is the enemy’s turn. Bot is rescued.
    """
    cascade = [
    ("""
    ########
    #ya.. b#
    #x     #
    ########
    """,{}),

    ("""
    ########
    #a .. b#
    #x     #
    ########
    """, {'y':(6,1)}),

    ("""
    ########
    #a ..by#
    #x     #
    ########
    """,{}),
    ]
    def move(bot, state):
        if bot.is_blue and bot.turn == 0 and bot.round == 1:
            return (1, 1)
        if bot.is_blue and bot.turn == 1 and bot.round == 1:
            return (5, 1)
        return bot.position
    layouts = [layout.parse_layout(l,bots=b) for l,b in cascade]
    state = setup_game([move, move], max_rounds=5, layout_dict=layouts[0])
    assert state['bots'] == layouts[0]['bots']
    state = game.play_turn(state) # Bot 0 moves, kills 3. Bot 2 and 3 are on same spot
    assert state['bots'] == layouts[1]['bots']
    state = game.play_turn(state) # Bot 1 stands. Bot 2 and 3 are on same spot
    assert state['bots'] == layouts[1]['bots']
    state = game.play_turn(state) # Bot 2 moves. Rescues itself
    assert state['bots'] == layouts[2]['bots']


def test_cascade_kill_rescue_2():
    """ Checks that killing occurs only for the bot whose turn it is
    or for any bot that this bot moves onto.
    If a bot respawns on an enemy, it will only be killed when it is its own
    or the enemy’s turn (and neither of them moves).
    If enemy moves before it is the bot’s turn. Bot is rescued.
    """
    cascade = [
    ("""
    ########
    #y ..  #
    #xa   b#
    ########
    """,{}),

    ("""
    ########
    #y ..  #
    #a    x#
    ########
    """, {'b':(6,2)}),

    ("""
    ########
    #y ..  #
    #a   xb#
    ########
    """, {}),
    ]
    def move(bot, state):
        if bot.is_blue and bot.turn == 0 and bot.round == 1:
            return (1, 2)
        if not bot.is_blue and bot.turn == 0 and bot.round == 1:
            return (5, 2)
        return bot.position
    layouts = [layout.parse_layout(l, bots=b) for l,b in cascade]
    state = setup_game([move, move], max_rounds=5, layout_dict=layouts[0])
    assert state['bots'] == layouts[0]['bots']
    state = game.play_turn(state) # Bot 0 moves, kills 1. Bot 1 and 2 are on same spot
    assert state['bots'] == layouts[1]['bots']
    state = game.play_turn(state) # Bot 1 moves. Bot 2 is rescued.
    assert state['bots'] == layouts[2]['bots']


def test_cascade_suicide():
    cascade = [
    ("""
    ########
    #x ..ay#
    #     b#
    ########
    """,{}),

    ("""
    ########
    #a .. y#
    #     b#
    ########
    """,{'x':(1,1)}),

    ("""
    ########
    #a .. y#
    #     x#
    ########
    """, {'b':(6,2)}),

    ("""
    ########
    #a .. y#
    #b    x#
    ########
    """,{}),
    ]
    def move(bot, state):
        if bot.is_blue and bot.turn == 0 and bot.round == 1:
            return (6, 1)
        return bot.position
    layouts = [layout.parse_layout(l, bots=b) for l,b in cascade]
    state = setup_game([move, move], max_rounds=5, layout_dict=layouts[0])
    assert state['bots'] == layouts[0]['bots']
    state = game.play_turn(state) # Bot 0 moves onto 3. Gets killed. Bot 0 and 1 are on same spot.
    assert state['bots'] == layouts[1]['bots']
    state = game.play_turn(state) # Bot 1 moves, gets killed. Bot 1 and 2 are on same spot
    assert state['bots'] == layouts[2]['bots']
    state = game.play_turn(state) # Bot 2 moves, gets killed.
    assert state['bots'] == layouts[3]['bots']


def test_moving_through_maze():
    test_start = """
        ######
        #a . #
        #.. x#
        #b  y#
        ###### """
    parsed = layout.parse_layout(test_start)
    teams = [
        stepping_player('>-v>>>-', '-^^->->'),
        stepping_player('<<-<<<-', '-------')
    ]
    state = setup_game(teams, layout_dict=parsed, max_rounds=8)

    # play first round
    for i in range(4):
        state = game.play_turn(state)
    test_first_round = layout.parse_layout(
        """ ######
            # a. #
            #..x #
            #b  y#
            ###### """)

    assert test_first_round['bots'] == state['bots']
    assert test_first_round['food'] == list(state['food'][0]) + list(state['food'][1])
    assert state['score'] == [0, 0]

    for i in range(4):
        state = game.play_turn(state)
    test_second_round = layout.parse_layout(
        """ ######
            # a. #
            #bx  #
            #   y#
            ###### """, bots={'b': (1, 2)}, food=[(1, 2)]) # b sitting on food

    assert test_second_round['bots'] == state['bots']
    assert test_second_round['food'] == list(state['food'][0]) + list(state['food'][1])
    assert state['score'] == [0, 1]

    for i in range(4):
        state = game.play_turn(state)
    test_third_round = layout.parse_layout(
        """ ######
            #b . #
            #.a x#
            #   y#
            ###### """)

    assert test_third_round['bots'] == state['bots']
    assert test_third_round['food'] == list(state['food'][0]) + list(state['food'][1])
    assert state['score'] == [game.KILL_POINTS, 1]

    for i in range(4):
        state = game.play_turn(state)
    test_fourth_round = layout.parse_layout(
        """ ######
            #b . #
            #a x #
            #   y#
            ###### """, bots={'a': (1, 2)}, food=[(1, 2)]) # a sitting on food

    assert test_fourth_round['bots'] == state['bots']
    assert test_fourth_round['food'] == list(state['food'][0]) + list(state['food'][1])
    assert state['score'] == [game.KILL_POINTS, game.KILL_POINTS + 1]

    for i in range(4):
        state = game.play_turn(state)
    test_fifth_round = layout.parse_layout(
        """ ######
            # b. #
            #.a x#
            #   y#
            ###### """)
    assert test_fifth_round['bots'] == state['bots']
    assert test_fifth_round['food'] == list(state['food'][0]) + list(state['food'][1])
    assert state['score'] == [game.KILL_POINTS * 2, game.KILL_POINTS + 1]

    for i in range(4):
        state = game.play_turn(state)
    test_sixth_round = layout.parse_layout(
        """
            ######
            # b. #
            #a x #
            #   y#
            ###### """, bots={'a': (1, 2)}, food=[(1, 2)]) # a sitting on food

    assert test_sixth_round['bots'] == state['bots']
    assert test_sixth_round['food'] == list(state['food'][0]) + list(state['food'][1])
    assert state['score'] == [game.KILL_POINTS * 2, game.KILL_POINTS * 2+ 1]

    for i in range(3): # !! Only move three bots
        state = game.play_turn(state)

    test_seventh_round = layout.parse_layout(
        """
            ######
            #  b #
            #a x #
            #   y#
            ###### """, bots={'a': (1, 2)}, food=[(1, 2)]) # a sitting on food

    assert test_seventh_round['bots'] == state['bots']
    assert test_seventh_round['food'] == list(state['food'][0]) + list(state['food'][1])
    assert state['score'] == [game.KILL_POINTS * 2 + 1, game.KILL_POINTS * 2 + 1]
    assert state['gameover'] == True
    assert state['whowins'] == 2

    with pytest.raises(ValueError):
        state = game.play_turn(state)


@pytest.mark.parametrize('score', ([[3, 3], 2], [[1, 13], 1], [[13, 1], 0]))
def test_play_turn_maxrounds(score):
    """Check that game quits at maxrounds and choses correct winner"""
    # this works for ties as well, because there are no points to be gained at init positions
    game_state = setup_random_basic_gamestate()
    game_state["round"] = 301
    game_state["score"] = score[0]
    game_state_new = game.play_turn(game_state)
    assert game_state_new["gameover"]
    assert game_state_new["whowins"] == score[1]

def test_play_turn_move():
    """Checks that bot is moved to intended space"""
    turn = 0
    parsed_l = layout.parse_layout(small_layout)
    game_state = {
        "food": parsed_l["food"],
        "walls": parsed_l["walls"],
        "bots": parsed_l["bots"],
        "shape": parsed_l["shape"],
        "max_rounds": 300,
        "team_names": ("a", "b"),
        "turn": turn,
        "round": 0,
        "timeout": [],
        "gameover": False,
        "whowins": None,
        "team_say": "bla",
        "score": 0,
        "error_limit": 5,
        "kills":[0]*4,
        "deaths": [0]*4,
        "bot_was_killed": [False]*4,
        "errors": [[], []],
        "fatal_errors": [{}, {}],
        "rnd": random.Random()
        }
    legal_positions = get_legal_positions(game_state["walls"], game_state["shape"], game_state["bots"][turn])
    game_state_new = apply_move(game_state, legal_positions[0])
    assert game_state_new["bots"][turn] == legal_positions[0]



def setup_random_basic_gamestate(*, round=1, turn=0):
    """helper function for testing play turn"""
    parsed_l = layout.parse_layout(small_layout)

    stopping = lambda bot, s: (bot.position, s)

    game_state = setup_game([stopping, stopping], layout_dict=parsed_l)
    game_state['round'] = round
    game_state['turn'] = turn
    return game_state


def setup_specific_basic_gamestate(round=0, turn=0):
    """helper function for testing play turn"""
    l = """
##################
#. ... .##.     y#
# # #  .  .### #x#
# # ##.   .      #
#      .   .## # #
#a# ###.  .  # # #
#b     .##. ... .#
##################
"""
    parsed_l = layout.parse_layout(l)

    stopping = lambda bot, s: (bot.position, s)

    game_state = setup_game([stopping, stopping], layout_dict=parsed_l)
    game_state['round'] = round
    game_state['turn'] = turn
    return game_state


def test_max_rounds():
    l = """
    ########
    #ba..xy#
    #      #
    ########
    """
    def move(bot, s):
        # in the first round (round #1),
        # all bots move to the south
        if bot.round == 1:
            # go one step to the right
            return (bot.position[0], bot.position[1] + 1)
        else:
            # There should not be more then one round in this test
            raise RuntimeError("We should not be here in this test")

    l = layout.parse_layout(l)
    assert l['bots'][0] == (2, 1)
    assert l['bots'][1] == (5, 1)
    assert l['bots'][2] == (1, 1)
    assert l['bots'][3] == (6, 1)
    # max_rounds == 1 should call move just once
    final_state = run_game([move, move], layout_dict=l, max_rounds=1)
    assert final_state['round'] == 1
    assert final_state['bots'][0] == (2, 2)
    assert final_state['bots'][1] == (5, 2)
    assert final_state['bots'][2] == (1, 2)
    assert final_state['bots'][3] == (6, 2)
    # max_rounds == 2 should finish and have the first team lose
    final_state = run_game([move, move], layout_dict=l, max_rounds=2)
    assert final_state['round'] == 2
    assert final_state['turn'] == 0
    assert final_state['bots'][0] == (2, 2)
    assert final_state['bots'][1] == (5, 2)
    assert final_state['bots'][2] == (1, 2)
    assert final_state['bots'][3] == (6, 2)
    assert final_state['gameover']
    assert final_state['whowins'] == 1
    assert final_state['fatal_errors'][0][0] == {
        'type': 'FatalException',
        'description': 'Exception in client (RuntimeError): We should not be here in this test',
        'round': 2,
        'turn': 0,
    }


def test_update_round_counter():
    tests = {
        (None, None): (1, 0),
        (1, 0): (1, 1),
        (1, 1): (1, 2),
        (1, 2): (1, 3),
        (1, 3): (2, 0),
        (2, 3): (3, 0)
    }

    for (round0, turn0), (round1, turn1) in tests.items():
        res = game.next_round_turn({'turn': turn0,
                                    'round': round0,
                                    'gameover': False,})
        assert all(item in res.items() for item in {'turn': turn1, 'round': round1}.items())

    for (round0, turn0), (round1, turn1) in tests.items():
        with pytest.raises(ValueError):
            res = game.next_round_turn({'turn': turn0,
                                        'round': round0,
                                        'gameover': True,})


def test_last_round_check():
    # (max_rounds, current_round, turn): gameover
    test_map = {
        (1, None, None): False,
        (1, 1, 0): False,
        (1, 1, 3): True,
    }
    for test_val, test_res in test_map.items():
        max_rounds, current_round, current_turn = test_val
        state = {
            'max_rounds': max_rounds,
            'round': current_round,
            'turn': current_turn,
            'error_limit': 5,
            'fatal_errors': [[],[]],
            'errors': [[],[]],
            'gameover': False,
            'score': [0, 0],
            'food': [{(1,1)}, {(1,1)}] # dummy food
        }
        res = game.check_gameover(state, detect_final_move=True)
        assert res['gameover'] == test_res


@pytest.mark.parametrize(
    'team_errors, team_wins', [
        (((0, 0), (0, 0)), False),
        (((0, 1), (0, 0)), False),
        (((0, 0), (0, 1)), False),
        (((0, 2), (0, 2)), False),
        (((0, 4), (0, 0)), False),
        (((0, 0), (0, 4)), False),
        (((0, 4), (0, 4)), False),
        (((0, 5), (0, 0)), 1),
        (((0, 0), (0, 5)), 0),
        (((0, 5), (0, 5)), 2),
        (((1, 0), (0, 0)), 1),
        (((0, 0), (1, 0)), 0),
        (((1, 0), (1, 0)), 2),
        (((1, 1), (1, 0)), 2),
        (((1, 0), (0, 5)), 1),
        (((0, 5), (1, 0)), 0),
    ]
)
def test_error_finishes_game(team_errors, team_wins):
    # the mapping is as follows:
    # [(num_fatal_0, num_errors_0), (num_fatal_1, num_errors_1), result_flag]
    # the result flag: 0/1: team 0/1 wins, 2: draw, False: no winner yet

    (fatal_0, errors_0), (fatal_1, errors_1) = team_errors
    # just faking a bunch of errors in our game state
    state = {
        "error_limit": 5,
        "fatal_errors": [[None] * fatal_0, [None] * fatal_1],
        "errors": [[None] * errors_0, [None] * errors_1]
    }
    res = game.check_gameover(state)
    if team_wins is False:
        assert res["whowins"] is None
        assert res["gameover"] is False
    else:
        assert res["whowins"] == team_wins
        assert res["gameover"] is True


@pytest.mark.parametrize('bot_to_move', [0, 1, 2, 3])
def test_finished_when_no_food(bot_to_move):
    """ Test that the game is over when a team has eaten its food. """
    l = """
    ########
    #  a.b #
    # y.x  #
    ########
    """
    bot_turn = bot_to_move // 2
    team_to_move = bot_to_move % 2
    def move(bot, s):
        if team_to_move == 0 and bot.is_blue and bot_turn == bot._bot_turn:
            return (4, 1)
            # eat the food between 0 and 2
        if team_to_move == 1 and (not bot.is_blue) and bot_turn == bot._bot_turn:
            # eat the food between 3 and 1
            return (3, 2)
        return bot.position

    l = layout.parse_layout(l)
    final_state = run_game([move, move], layout_dict=l, max_rounds=20)
    assert final_state['round'] == 1
    assert final_state['turn'] == bot_to_move


def test_minimal_game():
    def move(b, s):
        return b.position

    layout_name, layout_string = layout.get_random_layout()
    l = layout.parse_layout(layout_string)
    final_state = run_game([move, move], max_rounds=20, layout_dict=l)
    assert final_state['gameover'] is True
    assert final_state['score'] == [0, 0]
    assert final_state['round'] == 20

def test_minimal_losing_game_has_one_error():
    def move0(b, s):
        if b.round == 1 and b._bot_index == 0:
            # trigger a bad move in the first round
            return (0, 0)
        else:
            return b.position
    def move1(b, s):
        return b.position

    layout_name, layout_string = layout.get_random_layout()
    l = layout.parse_layout(layout_string)
    final_state = run_game([move0, move1], max_rounds=20, layout_dict=l)
    assert final_state['gameover'] is True
    assert final_state['score'] == [0, 0]
    assert len(final_state['errors'][0]) == 1
    assert len(final_state['errors'][1]) == 0
    assert final_state['round'] == 20


def test_minimal_remote_game():
    def move(b, s):
        return b.position

    layout_name, layout_string = layout.get_random_layout()
    l = layout.parse_layout(layout_string)
    final_state = run_game(["test/demo01_stopping.py", move], max_rounds=20, layout_dict=l)
    final_state = run_game(["test/demo01_stopping.py", 'test/demo02_random.py'], max_rounds=20, layout_dict=l)
    assert final_state['gameover'] is True
    assert final_state['score'] == [0, 0]
    assert final_state['round'] == 20


def test_non_existing_file():
    # TODO: Change error message to be more meaningful
    layout_name, layout_string = layout.get_random_layout()
    l = layout.parse_layout(layout_string)
    res = run_game(["blah", "nothing"], max_rounds=1, layout_dict=l)
    assert res['fatal_errors'][0][0] == {
        'description': '("Could not load blah: No module named \'blah\'", \'ModuleNotFoundError\')',
        'round': None,
        'turn': 0,
        'type': 'PlayerDisconnected'
    }

# TODO: Get it working again on Windows
@pytest.mark.skipif(_mswindows, reason="Test fails on some Python versions.")
def test_remote_errors(tmp_path):
    # TODO: Change error messages to be more meaningful

    syntax_error = FIXTURE_DIR / 'player_syntax_error'
    import_error = FIXTURE_DIR / 'player_import_error'

    layout_name, layout_string = layout.get_random_layout()
    l = layout.parse_layout(layout_string)

    res = run_game([str(syntax_error), str(import_error)], layout_dict=l, max_rounds=20)
    # Error messages have changed in Python 3.10. We can only do approximate maching
    assert "SyntaxError" in res['fatal_errors'][0][0].pop('description')
    assert res['fatal_errors'][0][0] == {
        'round': None,
        'turn': 0,
        'type': 'PlayerDisconnected'
    }
    # Both teams fail during setup: DRAW
    assert res['whowins'] == 2
    res = run_game(["0", str(import_error)], layout_dict=l, max_rounds=20)
    # Error messages have changed in Python 3.10. We can only do approximate maching
    assert "ModuleNotFoundError" in res['fatal_errors'][1][0].pop('description')
    assert res['fatal_errors'][1][0] == {
        'round': None,
        'turn': 1,
        'type': 'PlayerDisconnected'
    }
    assert res['whowins'] == 0
    res = run_game([str(import_error), "1"], layout_dict=l, max_rounds=20)
    # Error messages have changed in Python 3.10. We can only do approximate maching
    assert "ModuleNotFoundError" in res['fatal_errors'][0][0].pop('description')
    assert res['fatal_errors'][0][0] == {
        'round': None,
        'turn': 0,
        'type': 'PlayerDisconnected'
    }
    assert res['whowins'] == 1


@pytest.mark.parametrize('team_to_test', [0, 1])
def test_bad_move_function(team_to_test):
    """ Test that having a move function that returns a bad type
    appends a FatalException. """

    def stopping(b, state):
        return b.position
    def move0(b, state):
        return None
    def move1(b, state):
        return 0
    def move3(b, state):
        return 0, 0, 0
    def move4(b): # TypeError: move4() takes 1 positional argument but 2 were given
        return (0, 0), 0

    layout_name, layout_string = layout.get_random_layout()
    l = layout.parse_layout(layout_string)

    def test_run_game(move):
        # Flips the order of the teams depending on team_to_test
        if team_to_test == 0:
            teams = [move, stopping]
        elif team_to_test == 1:
            teams = [stopping, move]
        return run_game(teams, layout_dict=l, max_rounds=10)

    other = 1 - team_to_test

    res = test_run_game(move0)
    assert res['gameover']
    assert res['whowins'] == other
    assert res['fatal_errors'][team_to_test][0]['type'] == 'FatalException'
    assert res['fatal_errors'][team_to_test][0]['description'] == 'Exception in client (ValueError): Function move did not return a valid position: got None instead.'

    res = test_run_game(move1)
    assert res['gameover']
    assert res['whowins'] == other
    assert res['fatal_errors'][team_to_test][0]['type'] == 'FatalException'
    assert res['fatal_errors'][team_to_test][0]['description'] == 'Exception in client (ValueError): Function move did not return a valid position: got 0 instead.'

    res = test_run_game(move3)
    assert res['gameover']
    assert res['whowins'] == other
    assert res['fatal_errors'][team_to_test][0]['type'] == 'FatalException'
    assert res['fatal_errors'][team_to_test][0]['description'] == 'Exception in client (ValueError): Function move did not return a valid position: got (0, 0, 0) instead.'

    res = test_run_game(move4)
    assert res['gameover']
    assert res['whowins'] == other
    assert res['fatal_errors'][team_to_test][0]['type'] == 'FatalException'
    assert "takes 1 positional argument but 2 were given" in res['fatal_errors'][team_to_test][0]['description']


def test_setup_game_run_game_have_same_args():
    # We want to ensure that setup_game and run_game provide
    # the same API.

    # check that the parameters are the same
    params_setup_game = inspect.signature(setup_game).parameters.keys()
    params_run_game = inspect.signature(run_game).parameters.keys()
    assert params_setup_game == params_run_game

    # As run_game calls setup_game, we want to ensure that if a default is given
    # in both setup_game and run_game it has the same value in both lists.
    common_defaults = setup_game.__kwdefaults__.keys() & run_game.__kwdefaults__.keys()
    for kwarg in common_defaults:
        assert setup_game.__kwdefaults__[kwarg] == run_game.__kwdefaults__[kwarg], f"Default values for {kwarg} are different"


@pytest.mark.parametrize('bot_to_move', range(4))
# all combinations of True False in a list of 4
@pytest.mark.parametrize('bot_was_killed_flags', itertools.product(*[(True, False)] * 4))
def test_apply_move_resets_bot_was_killed(bot_to_move, bot_was_killed_flags):
    """ Check that `prepare_bot_state` sees the proper bot_was_killed flag
    and that `apply_move` will reset the flag to False. """
    team_id = bot_to_move % 2
    other_bot = (bot_to_move + 2) % 4
    other_team_id = 1 - team_id

    # specify which bot should move
    test_state = setup_random_basic_gamestate(turn=bot_to_move)

    bot_was_killed_flags = list(bot_was_killed_flags) # needs to be a list
    test_state['bot_was_killed'] = bot_was_killed_flags[:] # copy to avoid reference issues

    # create bot state for current turn
    current_bot_position = test_state['bots'][bot_to_move]
    bot_state = game.prepare_bot_state(test_state)

    # bot state should have proper bot_was_killed flag
    assert bot_state['team']['bot_was_killed'] == bot_was_killed_flags[team_id::2]

    # apply a dummy move that should reset bot_was_killed for the current bot
    new_test_state = game.apply_move(test_state, current_bot_position)

    # the bot_was_killed flag should be False again
    assert test_state['bot_was_killed'][bot_to_move] == False

    # the bot_was_killed flags for other bot should still be as before
    assert test_state['bot_was_killed'][other_bot] == bot_was_killed_flags[other_bot]

    # all bot_was_killed flags for other team should still be as before
    assert test_state['bot_was_killed'][other_team_id::2] == bot_was_killed_flags[other_team_id::2]


def test_bot_does_not_eat_own_food():
    test_layout = """
        ######
        #a .y#
        #.bx #
        ######
    """
    teams = [
        stepping_player('v', '<'),
        stepping_player('^', '<')
    ]
    state = setup_game(teams, layout_dict=layout.parse_layout(test_layout), max_rounds=2)
    assert state['bots'] == [(1, 1), (3, 2), (2, 2), (4, 1)]
    assert state['food'] == [{(1, 2)}, {(3, 1)}]
    for i in range(4):
        state = play_turn(state)
    assert state['bots'] == [(1, 2), (3, 1), (1, 2), (3, 1)]
    assert state['food'] == [{(1, 2)}, {(3, 1)}]


def test_suicide_win():
    # Test how a bot eats a food pellet that the enemy sits on
    # Since it is the last pellet, the game will end directly
    test_layout = """
        ######
        #a .x#
        #.   #
        #b  y#
        ######
    """
    teams = [
        stepping_player('>>', '--'),
        stepping_player('<-', '--')
    ]
    state = setup_game(teams, layout_dict=layout.parse_layout(test_layout), max_rounds=2)
    assert state['bots'] == [(1, 1), (4, 1), (1, 3), (4, 3)]
    assert state['food'] == [{(1, 2)}, {(3, 1)}]
    # play until finished
    state = run_game(teams, layout_dict=layout.parse_layout(test_layout), max_rounds=2)
    # bot 0 has been reset
    assert state['bots'] == [(1, 2), (3, 1), (1, 3), (4, 3)]
    assert state['food'] == [{(1, 2)}, set()]
    assert state['gameover'] == True
    assert state['whowins'] == 1
    assert state['round'] == 2
    assert state['turn'] == 0
    assert state['score'] == [1, game.KILL_POINTS]


def test_double_suicide():
    # Test how a bot can be killed when it runs into two bots
    test_layout = """
        ######
        # bx #
        #. y.#
        ######
    """
    teams = [
        stepping_player('-', '-'),
        stepping_player('<', '-')
    ]
    state = setup_game(teams, layout_dict=layout.parse_layout(test_layout, bots={'a':(2,1)}),
            max_rounds=2)
    assert state['bots'] == [(2, 1), (3, 1), (2, 1), (3, 2)]
    assert state['food'] == [{(1, 2)}, {(4, 2)}]
    # play a two turns so that 1 moves
    state = play_turn(state)
    state = play_turn(state)
    # bot 1 has been reset
    assert state['bots'] == [(2, 1), (4, 2), (2, 1), (3, 2)]
    assert state['food'] == [{(1, 2)}, {(4, 2)}]
    assert state['gameover'] == False
    assert state['round'] == 1
    assert state['turn'] == 1
    # only a single KILL_POINT has been given
    assert state['score'] == [game.KILL_POINTS, 0]


def test_remote_game_closes_players_on_exit():
    layout_name, layout_string = layout.get_random_layout()
    l = layout.parse_layout(layout_string)

    # run a remote demo game with "0" and "1"
    state = run_game(["0", "1"], layout_dict=l, max_rounds=20, allow_exceptions=True)
    assert state["gameover"]
    # Check that both processes have exited
    assert state["teams"][0].proc[0].wait(timeout=3) == 0
    assert state["teams"][1].proc[0].wait(timeout=3) == 0


def test_manual_remote_game_closes_players():
    layout_name, layout_string = layout.get_random_layout()
    l = layout.parse_layout(layout_string)

    # run a remote demo game with "0" and "1"
    state = setup_game(["0", "1"], layout_dict=l, max_rounds=10, allow_exceptions=True)
    assert not state["gameover"]
    while not state["gameover"]:
        # still running
        # still running
        assert state["teams"][0].proc[0].poll() is None
        assert state["teams"][1].proc[0].poll() is None
        state = play_turn(state)

    # Check that both processes have exited
    assert state["teams"][0].proc[0].wait(timeout=3) == 0
    assert state["teams"][1].proc[0].wait(timeout=3) == 0


def test_invalid_setup_game_closes_players():
    layout_name, layout_string = layout.get_random_layout()
    l = layout.parse_layout(layout_string)

    # setup a remote demo game with "0" and "1" but bad max rounds
    state = setup_game(["0", "1"], layout_dict=l, max_rounds=0, allow_exceptions=True)
    assert state["gameover"]
    # Check that both processes have exited
    assert state["teams"][0].proc[0].wait(timeout=3) == 0
    assert state["teams"][1].proc[0].wait(timeout=3) == 0


@pytest.mark.parametrize('move_request, expected_prev, expected_req, expected_success', [
    # invalid moves (wrong types)
    # invalid moves end the game immediately, but the viewer will still interpret
    # the outcome of requested_moves, which is the reason we should test this
    (None,          (1, 1), None, False),
    (-1,            (1, 1), None, False),
    ((-1,),         (1, 1), None, False),
    ((-1, 1, 2),    (1, 1), None, False),
    ("a" ,          (1, 1), None, False),
    # not in maze
    ((-1, 2),       (1, 1), (-1, 2), False),
    # on wall
    ((2, 1),        (1, 1), (2, 1), False),
    # good move
    ((1, 2),        (1, 1), (1, 2), True),
    ]
)
def test_requested_moves(move_request, expected_prev, expected_req, expected_success):
    # test the possible return values of gamestate['requested_moves']
    test_layout = """
        ######
        #a#.y#
        #.bx #
        ######
    """
    def move(bot, state):
        return move_request
    teams = [
        move,
        stopping_player
    ]
    state = setup_game(teams, layout_dict=layout.parse_layout(test_layout), max_rounds=2)
    assert state['requested_moves'] == [None, None, None, None]
    state = play_turn(state)
    assert state['requested_moves'][1:] == [None, None, None]
    assert state['requested_moves'][0] == {'previous_position': (1, 1), 'requested_position': expected_req, 'success': expected_success}



#!/usr/bin/env python3

from pelita.layout import parse_layout
from pelita.game import run_game
from pelita.player import smart_eating_player, smart_random_player

if __name__ == '__main__':
    layout = parse_layout(
        """ ##################
            #a#.  .  # .     #
            #b#####    #####y#
            #     . #  .  .#x#
            ################## """)

    teams = [smart_eating_player, smart_random_player]
    run_game(teams, layout_dict=layout, max_rounds=200, viewers=['ascii'])
